from django.contrib.auth.models import User, Group
from django.db.models import Q, Count
from django.shortcuts import get_object_or_404
from django.http import HttpResponseRedirect
from django.core.urlresolvers import reverse

from racks.inventory.models import InventoryObject, Lab, InventoryObjectLabel
from racks import crud
from django.views.generic.base import ContextMixin
from itertools import chain


class GetAccountMixin(ContextMixin):
    def get_object(self, **kwargs):
        return get_object_or_404(User, username=kwargs['username'])


class AccountsListView(crud.ListView):
    def get_queryset(self):
        return self.model.objects.all().annotate(
            inventory_objects_count=Count('assigned_objects'),
            labs_count=Count('assigned_labs'),
        ).order_by('-inventory_objects_count', 'username')


class AccountDetailView(GetAccountMixin, crud.DetailView):
    template = 'profile'


class AccountAssignedDetailView(GetAccountMixin, crud.DetailView):
    template = 'assigned'

    def get_context_data(self, **kwargs):
        context = super(AccountAssignedDetailView, self).get_context_data(**kwargs)
        user = self.get_object(**kwargs)
        user_groups = user.groups.all()

        context.update({
            'operations_allowed': self.request.user.is_superuser or self.request.user == user,
            'user': user,
            'assigned_objects':
                InventoryObject.objects.filter(
                    Q(assigned_user=user) | Q(assigned_group__in=user_groups)
                ).distinct().prefetch_related('labels', 'dcim'),
            'assigned_labs':
                Lab.objects.filter(
                    Q(assigned_user=user) | Q(assigned_group__in=user_groups)
                ).distinct(),
        })
        return context


class AccountBulkActionsView(crud.CRUDView):
    template = 'bulk_actions'

    def view_allowed(self, request, *args, **kwargs):
        user = get_object_or_404(User, **kwargs)
        return request.user.is_superuser or request.user == user

    def post(self, request, *args, **kwargs):
        context = self.get_context_data(**kwargs)

        user = get_object_or_404(User, **kwargs)

        if request.POST.get('action_assign'):
            action = 'assign'
        elif request.POST.get('action_labels'):
            action = 'labels'
        else:
            action = request.POST.get('action')

        # Query for selected objects
        inventory = InventoryObject.objects.filter(
            pk__in=request.POST.getlist('io')
        ).select_related('assigned_user', 'assigned_group').prefetch_related('labels')
        labs = Lab.objects.filter(
            pk__in=request.POST.getlist('lab')
        ).select_related('assigned_user', 'assigned_group')

        # Redirect back is nothing is selected
        if not(inventory or labs):
            return HttpResponseRedirect(reverse(self.view_set.url_names['assigned'], kwargs=kwargs))

        # Generate resulting context for selected objects
        inventory = filter(lambda obj: obj.assigned_user == user, inventory)
        io_list = map(lambda obj: obj.pk, inventory)
        labs = filter(lambda obj: obj.assigned_user == user, labs)
        lab_list = map(lambda obj: obj.pk, labs)
        labels = sorted(set(chain(*[io.labels.all() for io in inventory])), key=lambda l: l.name)

        # Validate selected actions
        target = request.POST.get('target', '')
        recipient = crud.search_users_and_groups_by_name(target)
        add_label = map(lambda l: l.strip(), request.POST.getlist('add_label'))
        del_label = request.POST.getlist('del_label')

        perform_actions = bool(target or add_label or del_label)

        if target:
            # If reassignment is selected, we only proceed once we get a 100% name match
            perform_actions = False
            if isinstance(recipient, User):
                perform_actions |= (recipient.username == target)
                target = recipient.username
            elif isinstance(recipient, Group):
                perform_actions |= (recipient.name == target)
                target = recipient.name

        # If everything is ok, perform actions and return to assigned page.
        if perform_actions:
            if recipient:
                to_user = recipient if isinstance(recipient, User) else None
                to_group = recipient if isinstance(recipient, Group) else None
                update = {}
                if to_user:
                    update['assigned_user'] = to_user
                if to_group:
                    update['assigned_group'] = to_group
                InventoryObject.objects.filter(pk__in=io_list).update(**update)
                Lab.objects.filter(pk__in=lab_list).update(**update)
            for label_name in add_label:
                label, created = InventoryObjectLabel.objects.get_or_create(name=label_name)
                for io in inventory:
                    io.labels.add(label)
            for label_name in del_label:
                try:
                    label = InventoryObjectLabel.objects.get(name=label_name)
                except InventoryObjectLabel.DoesNotExist:
                    continue
                for io in inventory:
                    io.labels.remove(label)
            return HttpResponseRedirect(reverse(self.view_set.url_names['assigned'], kwargs=kwargs))

        context.update(
            user=user, action=action, add_label=add_label, del_label=del_label,
            inventory=inventory, io_list=io_list, labs=labs, lab_list=lab_list, labels=labels,
            target=target, recipient=recipient, multiple_matches=isinstance(recipient, list),
        )

        return self.render_to_response(context)


class AccountViewSet(crud.ViewSet):
    PATTERNS = (
        (r'^$', 'list'),
        (r'^(?P<username>[0-9A-Za-z_@\-\+\.]{1,30})/$', 'profile'),
        (r'^(?P<username>[0-9A-Za-z_@\-\+\.]{1,30})/assigned$', 'assigned'),
        (r'^(?P<username>[0-9A-Za-z_@\-\+\.]{1,30})/actions', 'actions'),
    )

    model = User
    list_view = AccountsListView
    profile_view = AccountDetailView
    assigned_view = AccountAssignedDetailView
    actions_view = AccountBulkActionsView
    fields = '__all__'


urlpatterns = AccountViewSet('account').urlpatterns
