from django import forms
from django.conf.urls import url
from django.contrib.auth.decorators import login_required
from django.contrib.auth.models import User, Group
from django.core.exceptions import ImproperlyConfigured
from django.core.paginator import Paginator
from django.core.urlresolvers import reverse
from django.forms.forms import BoundField
from django.forms.models import modelform_factory
from django.forms import widgets
from django.http import HttpResponseForbidden, HttpResponseRedirect
from django.shortcuts import get_object_or_404
from django.views.generic.base import TemplateView
from django.utils.decorators import method_decorator, classonlymethod
from django.utils.safestring import mark_safe

from racks.widgets import FilteredSelectMultiple
from guardian.shortcuts import (assign_perm, remove_perm, get_perms_for_model,
                                get_users_with_perms, get_groups_with_perms,
                                get_objects_for_user)
from racks.security import get_perms

import logging
from copy import deepcopy
from itertools import count, izip


log = logging.getLogger(__name__)


class CRUDView(TemplateView):
    """
    CRUDView is the basic class used in Racks CRUD layer. It provides a basis for a single view
    within a CRUD ViewSet.

    Most of the time, developer will inherit not from CRUDView directly, but from one of it's
    standardized descendants, which provide support for Racks security and permissions system and
    allow for rapid development of model-related view groups.
    """

    view_set = None
    model = None
    template = None

    @property
    def url_names(self):
        return self.view_set.url_names

    def get_model_verbose_name(self):
        return self.model._meta.verbose_name

    @method_decorator(login_required)
    def dispatch(self, request, *args, **kwargs):
        if self.view_allowed(request, *args, **kwargs):
            return super(CRUDView, self).dispatch(request, *args, **kwargs)
        else:
            return HttpResponseForbidden()

    @classonlymethod
    def as_view(cls, view_set):
        return super(CRUDView, cls).as_view(view_set=view_set, model=view_set.model)

    def get_template_names(self):
        if self.template_name:
            return [self.template_name]

        path = self.view_set.prefix.replace('.', '/')
        if not self.template:
            raise ImproperlyConfigured("'template_name' or 'template' variable should be defined")
        template = self.template

        return [
            '{}/{}.html'.format(path, template),
            'crud/{}.html'.format(template),
        ]

    def view_allowed(self, *args, **kwargs):
        return True

    def get_queryset(self):
        return self.model.objects.all()


class FieldForDisplay(object):
    def __init__(self, label):
        self.label = label


class ReadOnlyBoundField(BoundField):
    def as_widget(self, widget=None, attrs=None, only_initial=False):
        attrs = attrs or {}
        attrs['readonly'] = 'readonly'

        widget = widget or self.field.widget

        # Disable opening of Select fields. We're checking for type, because our
        # FilteredSelectMultiple widget is subclass of Select, but correctly handles readonly
        # attribute.
        if type(widget) is widgets.Select:
            attrs['disabled'] = True

        return super(ReadOnlyBoundField, self).as_widget(widget, attrs, only_initial)


class CRUDFormView(CRUDView):
    """
    CRUDFormView is the base for CRUD form classes, namely UpdateView and CreateView. It automates
    creation of model forms and has embedded support for Racks permissions system.
    """

    def get_initial(self, obj):
        initial = {}

        for field_name in self.view_set.related_foreign_key_fields:
            if obj:
                initial[field_name] = getattr(obj, field_name).all()

        return initial

    def get_form_class(self):
        if self.view_set.change_form_class:
            return self.view_set.change_form_class

        widgets = {
            field: FilteredSelectMultiple
            for field in self.view_set.filtered_select_fields
        }

        if self.view_set.widgets:
            widgets.update(self.view_set.widgets)

        return modelform_factory(self.model, fields=self.view_set.fields, widgets=widgets,
                                 formfield_callback=self.formfield_callback)

    def get_form(self, obj=None, data=None):
        form_class = self.get_form_class()
        form = form_class(data=data, instance=obj, initial=self.get_initial(obj))

        for field_name in self.view_set.related_foreign_key_fields:
            field = getattr(self.model, field_name)
            qs = field.related.model.objects.all()
            form.fields[field_name] = forms.ModelMultipleChoiceField(
                qs, required=False, widget=FilteredSelectMultiple
            )
        return form

    def formfield_callback(self, f, **kwargs):
        field_class = self.view_set.field_classes.get(f.name)

        if field_class:
            return field_class(**kwargs)

        return f.formfield(**kwargs)

    def get_success_url(self, obj):
        try:
            return obj.get_absolute_url()
        except AttributeError:
            return reverse(self.url_names['list'])

    def post(self, request, **kwargs):
        context = self.get_context_data(**kwargs)
        form = context['form']

        if form.is_valid():
            obj = form.save()
            return HttpResponseRedirect(self.get_success_url(obj))

        return self.render_to_response(context)


class ListView(CRUDView):
    paginate_by = 20
    template = 'change_list'
    display_actions = True  # Display Actions column or not

    def get_parameters(self):
        get_parameters = self.request.GET.copy()
        get_parameters.pop('page', None)
        return get_parameters

    def prefetch_permissions(self):
        user = self.request.user
        queryset = self.get_queryset()
        model = queryset.model
        perm_suffix = model.__name__.lower()
        edit_perm = get_objects_for_user(user, 'change_' + perm_suffix, queryset)
        delete_perm = get_objects_for_user(user, 'delete_' + perm_suffix, queryset)

        field_names = model._meta.get_all_field_names()
        owned_objects = None
        if 'assigned_user' in field_names:
            owned_objects = self.model.objects.filter(assigned_user=user)

        if 'assigned_group' in field_names:
            owned_objects |= self.model.objects.filter(assigned_group__user=user)

        pk_name = model._meta.pk.name

        can_edit = list(map(lambda x: x[pk_name], edit_perm.values(pk_name)))
        can_delete = list(map(lambda x: x[pk_name], delete_perm.values(pk_name)))

        # Optimization: this data can be moved to memcached, because on large DB it can take
        # more than 1 second to pull all objects which is owned directly by user or
        # indirectly via group
        owned = list(map(lambda x: x['id'], owned_objects.values('id'))) if owned_objects else []

        return {
            'can_edit': set(can_edit + owned),
            'can_delete': set(can_delete + owned),
        }

    def get_context_data(self, **kwargs):
        context = super(ListView, self).get_context_data(**kwargs)
        try:
            per_page = int(self.request.GET.get('per_page', self.paginate_by))
        except ValueError:
            per_page = self.paginate_by

        try:
            page_no = int(self.request.GET.get('page', 1))
        except ValueError:
            page_no = 1

        queryset = self.get_queryset()
        paginator = Paginator(queryset, per_page)
        page = paginator.page(page_no)

        context.update({
            'page': page,
        })
        context.update(self.prefetch_permissions())
        return context


class DetailView(CRUDView):
    template = 'detail'

    def check_perm(self, perm_name):
        return perm_name in get_perms(self.request.user, self.object)

    def get_object(self, **kwargs):
        return get_object_or_404(self.get_queryset(), pk=kwargs['pk'])

    def get_context_data(self, **kwargs):
        context = super(DetailView, self).get_context_data(**kwargs)
        context['object'] = self.object = self.get_object(**kwargs)
        return context


class UpdateView(CRUDFormView):
    template_name = 'crud/update_form.html'

    def view_allowed(self, request, *args, **kwargs):
        queryset = self.get_queryset()
        obj = get_object_or_404(queryset, pk=kwargs['pk'])
        return 'change_' + queryset.model.__name__.lower() in get_perms(request.user, obj)

    def get_context_data(self, **kwargs):
        context = super(UpdateView, self).get_context_data(**kwargs)

        context['object'] = obj = get_object_or_404(self.get_queryset(), pk=kwargs['pk'])

        if self.request.method == 'POST':
            context['form'] = self.get_form(obj=obj, data=self.request.POST)
        else:
            context['form'] = self.get_form(obj=obj)

        return context


class CreateView(CRUDFormView):
    template_name = 'crud/create_form.html'

    def view_allowed(self, request, *args, **kwargs):
        return 'add_' + self.model.__name__.lower() in get_perms(request.user, self.model())

    def get_context_data(self, **kwargs):
        context = super(CreateView, self).get_context_data(**kwargs)

        if self.request.method == 'POST':
            context['form'] = self.get_form(data=self.request.POST)
        else:
            context['form'] = self.get_form()

        return context


class DeleteView(CRUDView):
    template_name = 'crud/confirm_delete.html'

    def view_allowed(self, request, *args, **kwargs):
        queryset = self.get_queryset()
        obj = get_object_or_404(queryset, pk=kwargs['pk'])
        return 'delete_' + queryset.model.__name__.lower() in get_perms(request.user, obj)

    def get_success_url(self):
        return self.request.GET.get('next') or reverse(self.router.list_view.url_name)

    def delete_object(self, obj):
        """Actually performs object deletion. To be overridden in child classes if necessary."""
        obj.delete()

    def post(self, request, pk):
        obj = get_object_or_404(self.get_queryset(), pk=pk)
        self.delete_object(obj)
        return HttpResponseRedirect(self.get_success_url())


class AccessFormCheckbox(object):
    """
    Helper class for AccessView.

    Represents a single checkbox on user/object permission management form.
    """

    def __init__(self, index, name, checked, readonly):
        self.index = index
        self.name = name
        self.checked = ' checked="checked"' if checked else ''
        self.readonly = ' disabled' if readonly else ''

    def html(self):
        return mark_safe(
            '<input type="checkbox" id="id_{name}" name="{name}" '
            'value="{index}"{checked}{readonly} />'.format(**self.__dict__)
        )


class AccessFormRow(object):
    """
    Helper class for AccessView.

    Represents a single row of checkboxes on user/object permission management form, i.e. a set of
    permissions for a specific user or group.
    """

    def __init__(self, obj, index, applicant, permissions, **kwargs):
        model_name = obj._meta.object_name.lower()
        self.index = index
        self.applicant = applicant
        self.perms = []
        self.error = kwargs.get('error', None)
        self.options = kwargs.get('options', None)
        self.row_name = kwargs.get('row_name', '')
        if 'row_type' in kwargs:
            self.row_type = kwargs['row_type']
        elif isinstance(applicant, User):
            self.row_type = 'user'
        elif isinstance(applicant, Group):
            self.row_type = 'group'
        else:
            self.row_type = 'text'
        disable_manage = kwargs.get('disable_manage', True)
        applicant_perms = kwargs.get('applicant_perms', set())
        for perm in permissions:
            self.perms.append(AccessFormCheckbox(
                self.index,
                perm.codename,
                perm.codename in applicant_perms,
                disable_manage and (perm.codename == 'manage_%s' % model_name),
            ))


def search_users_and_groups_by_name(name, prioritize_group=False):
    """
    Search User and Group models for specific name or partial name.

    Function will attempt to search for an exact User match, then an exact Group match, and finally
    will try a partial match for all records in User and Group tables.

    Optional boolean parameter tells the function to prioritize exact Group match over User.

    Function will return a single User or Group record if an exact match is found or if partial
    match only returned a single record, None if nothing is found, or a list of User and Group
    records if there were multiple matches.

    :param name: name or partial name to search for
    :param prioritize_group: search for exact Group name match before trying User
    :return: one of: None, User instance, Group instance, list of User and Group instances
    """

    if not name:
        return None

    # Attempt for an exact name or username match:
    if prioritize_group:
        try:
            return Group.objects.get(name=name)
        except Group.DoesNotExist:
            pass
    try:
        return User.objects.get(username=name)
    except User.DoesNotExist:
        pass
    if not prioritize_group:
        try:
            return Group.objects.get(name=name)
        except Group.DoesNotExist:
            pass

    # Search all records for partial matches since exact match was not found:
    users = User.objects.filter(username__icontains=name).order_by('username')
    groups = Group.objects.filter(name__icontains=name).order_by('name')
    total_found = len(users) + len(groups)
    if total_found == 0:
        return None
    elif total_found == 1:
        if len(users):
            return users[0]
        else:
            return groups[0]
    else:
        return [(u.username, u.username) for u in users] + [(g.name, g.name) for g in groups]


class AccessForm(object):
    """
    Standardized permissions management form for Racks classes using 'security.permissions'.

    Designed to be model-agnostic, this view will use information available in the model itself
    to identify permissions that need to be managed. By default it will manage all permissions
    with the exception of 'add_modelname' which is a model-level permission instead of object-level.
    """

    def __init__(self, obj, current_user, **kwargs):
        self.object = obj
        self.user = current_user
        self.permissions = []
        # Starting values
        self.initial = []
        # Values received from form
        self.rows = []
        self.block_manage = \
            'manage_' + obj.__class__.__name__.lower() not in get_perms(current_user, obj)
        self.empty_rows = kwargs.get('empty_rows', 1)

    def set_initial(self):
        self.initial = []

        model_name = self.object.__class__.__name__.lower()
        manage = None
        self.permissions = []
        for perm in get_perms_for_model(self.object):
            if perm.codename == 'add_%s' % model_name:
                continue
            if perm.codename == 'manage_%s' % model_name:
                manage = perm
            else:
                self.permissions.append(perm)
        if manage is not None:
            self.permissions.append(manage)

        users = sorted(
            get_users_with_perms(self.object, True, False, False).items(),
            key=lambda u: u[0].username
        )
        for user, user_perms in users:
            self.initial.append(
                AccessFormRow(
                    self.object, len(self.initial), user, self.permissions,
                    applicant_perms=user_perms, disable_manage=self.block_manage,
                    row_name=user.username, row_type='user',
                )
            )

        groups = sorted(
            get_groups_with_perms(self.object, True).items(),
            key=lambda g: g[0].name
        )
        for group, group_perms in groups:
            self.initial.append(
                AccessFormRow(
                    self.object, len(self.initial), group, self.permissions,
                    applicant_perms=group_perms, disable_manage=self.block_manage,
                    row_name=group.name, row_type='group',
                )
            )

        self.rows = deepcopy(self.initial)

    def update_from_post(self, data):
        self.rows = []
        names = data.getlist('name')
        types = data.getlist('type')
        for index, row_name, row_type in izip(count(), names, types):
            if not row_name:
                continue
            error = None
            options = None
            match = search_users_and_groups_by_name(row_name, row_type == 'group')
            if match is None:
                error = 'No matches found'
                row_type = 'text'
            elif isinstance(match, list):
                error = 'Several matches found'
                options = [(row_name, 'Please select:'), ('', '(remove)')] + match
            else:
                row_type = 'user' if isinstance(match, User) else 'group'
            row_permissions = [
                p.codename for p in self.permissions if str(index) in data.getlist(p.codename)
            ]
            self.rows.append(AccessFormRow(
                self.object, len(self.rows), match, self.permissions, error=error,
                options=options, row_type=row_type, row_name=row_name,
                disable_manage=self.block_manage, applicant_perms=row_permissions
            ))

    def add_empty_rows(self):
        for index in xrange(self.empty_rows):
            self.rows.append(
                AccessFormRow(self.object, len(self.rows), None, self.permissions,
                              disable_manage=self.block_manage)
            )

    def is_valid(self):
        return all(not row.error for row in self.rows)

    def save(self):
        for row in self.rows:
            for perm in row.perms:
                if perm.readonly:
                    continue
                if perm.checked:
                    assign_perm(perm.name, row.applicant, self.object)
                else:
                    remove_perm(perm.name, row.applicant, self.object)


class AccessView(CRUDView):
    """
    CRUD view for access management form.
    """

    template_name = 'crud/access_form.html'

    def view_allowed(self, request, *args, **kwargs):
        queryset = self.get_queryset()
        obj = get_object_or_404(queryset, pk=kwargs['pk'])
        return 'manage_' + queryset.model.__name__.lower() in get_perms(request.user, obj)

    def get_success_url(self, obj):
        try:
            return obj.get_absolute_url()
        except AttributeError:
            return reverse(self.url_names['list'])

    def get_context_data(self, **kwargs):
        context = super(AccessView, self).get_context_data(**kwargs)

        context['object'] = self.object
        context['form'] = self.form

        return context

    def get(self, request, *args, **kwargs):
        self.object = get_object_or_404(self.get_queryset(), pk=kwargs['pk'])
        self.form = AccessForm(self.object, request.user)
        self.form.set_initial()
        self.form.add_empty_rows()
        context = self.get_context_data(**kwargs)
        return self.render_to_response(context)

    def post(self, request, *args, **kwargs):
        self.object = get_object_or_404(self.get_queryset(), pk=kwargs['pk'])
        self.form = AccessForm(self.object, request.user)
        self.form.set_initial()
        self.form.update_from_post(request.POST)
        if self.form.is_valid():
            self.form.save()
            return HttpResponseRedirect(self.get_success_url(self.object))
        else:
            self.form.add_empty_rows()
            context = self.get_context_data(**kwargs)
            return self.render_to_response(context)


class ViewSet(object):
    """
    A single set of views for Racks project.

    CRUD ViewSet allows grouping of several related views with a common root, usually working with
    the same model. By default, ViewSet provides support for paginated list view, read-only details
    view, create/edit views with dynamically generated forms, and permissions management view.

    Permissions management view is automatically enabled for models which have 'manage_modelname'
    permission defined. It can be enabled for other models as well by declaring explicitly in
    your ViewSet class:

        access_view = racks.crud.AccessView

    Alternatively, it can be disabled by setting `disable_access_view` class parameter to True.
    """

    PATTERNS = (
        (r'^$', 'list'),
        (r'^(?P<pk>[0-9]+)$', 'detail'),
        (r'^new$', 'create'),
        (r'^(?P<pk>[0-9]+)/edit$', 'update'),
        (r'^(?P<pk>[0-9]+)/delete$', 'delete'),
        (r'^(?P<pk>[0-9]+)/access$', 'access'),
    )

    list_view = ListView
    detail_view = None
    create_view = CreateView
    update_view = UpdateView
    delete_view = DeleteView
    access_view = None

    change_form_class = None
    filtered_select_fields = ()
    readonly_fields = ()
    related_foreign_key_fields = ()

    model = None
    fields = None
    field_classes = {}
    widgets = None

    disable_access_view = False

    def __init__(self, prefix):
        self.prefix = prefix
        self.url_names = {}
        self.urlpatterns = []

        if not self.disable_access_view and self.model:
            access_permission = 'manage_%s' % self.model._meta.object_name.lower()
            for perm in get_perms_for_model(self.model):
                if perm.codename == access_permission:
                    self.access_view = AccessView
                    break

        for pattern, name in self.PATTERNS:
            view_attr = name + '_view'
            view_class = getattr(self, view_attr)

            if view_class:
                url_name = '%s.%s' % (prefix, name)
                self.url_names[name] = url_name
                self.urlpatterns.append(url(pattern, view_class.as_view(self), name=url_name))

            self.url_names['object'] = self.url_names.get('detail') or self.url_names.get('update')
