import six
import logging

import requests
from django.db import IntegrityError, transaction
from django.conf import settings

from racks.dcim.models import Switch, SwitchPort, DCIM_EMPTY, DCIM_SERVER, DCIM_SWITCH, DCIM_OTHER
from racks.dcim.models import Device
from racks.inventory.models import InventoryObject


logger = logging.getLogger('racks.importers.dcim')  # Use importers logger instead of dcim one


class Session(object):
    def __init__(self):
        self.session = requests.session()
        self.session.auth = (settings.DCIM_USER, settings.DCIM_PASSWORD)

    def request(self, object_type, object_id=None, filters={}):
        if object_id is not None:
            if filters:
                # Code error, will be obvious from traceback
                raise TypeError('Either pass object_id or filters')
            path = '{}/{}'.format(object_type, object_id)
        else:
            path = object_type

        response = self.session.get(settings.DCIM_API + path, params=filters,
                                    timeout=settings.DCIM_API_TIMEOUT)

        assert response.status_code == 200

        # DCIM has a bug, it returns empty list instead of empty dict, handling this case
        return response.json()[object_type] or {}


def import_ports(switch, session, server_dcim_ids, switch_dcim_ids):
    deviceports = session.request('deviceport', switch.dcim_id)

    old_ports = {
        p.number: p
        for p in switch.ports.all()
    }

    for number, deviceport in deviceports.items():
        try:
            port = old_ports.pop(number)
        except KeyError:
            port = SwitchPort(number=number, switch=switch)

        port.label = deviceport['Label']

        if deviceport['ConnectedDeviceID'] is not None:
            port.dcim_server_id = int(deviceport['ConnectedDeviceID'])
            if port.dcim_server_id in server_dcim_ids:
                port.dcim_device_type = DCIM_SERVER
            elif port.dcim_server_id in switch_dcim_ids:
                port.dcim_device_type = DCIM_SWITCH
            else:
                port.dcim_device_type = DCIM_OTHER
        else:
            port.dcim_server_id = None
            port.dcim_device_type = DCIM_EMPTY

        port.save()

    # Delete ports in database but not in DCIM
    for port in old_ports.values():
        port.delete()


def short_username(full_name):
    if isinstance(full_name, six.string_types) and ' ' in full_name:
        first_name, last_name = full_name.split(' ', 1)
        first_name_letter = first_name[0] if len(first_name) > 0 else ''
        reporter_short = (first_name_letter + last_name.replace(' ', '')).lower()
    elif isinstance(full_name, six.string_types):
        reporter_short = full_name.lower()
    else:
        reporter_short = full_name
    return reporter_short


@transaction.atomic
def import_switches():

    server_dcim_ids = {
        row['dcim_id']
        for row in InventoryObject.objects.filter(dcim_id__isnull=False).values('dcim_id')
    }
    switch_dcim_ids = set()

    session = Session()

    devices = session.request('device', filters={
        'DeviceType': 'Switch',
    })

    old_switches = {
        s.dcim_id: s
        for s in Switch.objects.all()
    }
    actual_switches = []

    for device in devices:
        dcim_id = device['DeviceID']
        switch_dcim_ids.add(int(dcim_id))

        try:
            switch = old_switches.pop(dcim_id)
        except KeyError:
            switch = Switch(dcim_id=dcim_id)

        switch.label = device['Label']
        switch.management_address = device['PrimaryIP']

        if not switch.management_address:
            logger.error("Switch '%s' does not have valid PrimaryIP field", switch.label)
            continue

        try:
            switch.save()
        except IntegrityError:
            logger.exception("Switch id: '%s', label: '%s' save to DB failed",
                             device['DeviceID'], device['Label'])
            continue

        actual_switches.append(switch)

    for switch in actual_switches:
        try:
            import_ports(switch, session, server_dcim_ids, switch_dcim_ids)
        except:
            logger.exception("Ports import failed for switch %s", switch)
            raise

    # Delete switches in database but not in DCIM
    for s in old_switches.values():
        s.delete()


@transaction.atomic
def import_servers():
    filters = {'DeviceType': 'Server'}
    owners = settings.DCIM_SERVER_OWNER_IDS_LIST
    existing_servers = {
        s.dcim_id: s
        for s in InventoryObject.objects.exclude(dcim=None).select_related('dcim')
    }

    session = Session()

    # if we have owners apply filter, otherwise fetch all devices
    devices = session.request('device', filters=filters)

    for device in devices:
        dcim_id = device['DeviceID']
        dcim = Device.objects.update_or_create(pk=dcim_id, defaults={
            'device_type': device['DeviceType'],
            # label can't be null but IT guys have entries without label
            'label': device['Label'],
            'install_date': device['InstallDate'],
            'owner': device['Owner'],
            'notes': device['Notes'],
            'custom_values': device['CustomValues'] or {},
        })[0]

        try:
            server = existing_servers.pop(dcim_id)
        except KeyError:
            if not owners or device['Owner'] in owners:
                server = InventoryObject(dcim=dcim, source={'DCIM': dcim_id})
            else:
                continue

        # If no user or group are assigned to this server, we try to auto-assign it to reporter.
        if dcim.contact_person and server.assigned_user is None and server.assigned_group is None:
            logger.info("Assigning server %s to %s", server, dcim.contact_person)
            server.assigned_user = dcim.contact_person
        server.save()

        if dcim.jira_issue:
            server.attributes.update_or_create(name='DCIM Jira Issue',
                                               defaults={'value': dcim.jira_issue})

    for server in existing_servers.values():
        if server.name:
            logger.info("Delete DCIM entry %s as no longer exists in DCIM", repr(server.dcim))
        else:
            logger.info("Delete DCIM entry %s and Inventory Object %s as no longer exists in DCIM",
                        repr(server.dcim), repr(server))
            server.delete()
        server.dcim.delete()
