import six

from django.conf import settings
from django.core.exceptions import ImproperlyConfigured
from django.db import models
from django.dispatch import receiver
from django.utils.functional import cached_property
from django.utils.http import urlencode
from django.contrib.auth.models import User
from django.utils.encoding import python_2_unicode_compatible
import jsonfield

from racks.models import RacksModel

DCIM_EMPTY = 0
DCIM_SERVER = 1
DCIM_SWITCH = 4
DCIM_OTHER = -1

# Only DCIM_SERVER and DCIM_SWITCH are currently imported. All others are mapped to DCIM_OTHER.
DCIM_DEVICE_TYPES = (
    (DCIM_EMPTY, 'Empty'),
    (DCIM_SERVER, 'Server'),
    (2, 'Appliance'),
    (3, 'Storage Array'),
    (DCIM_SWITCH, 'Switch'),
    (5, 'Chassis'),
    (6, 'Patch Panel'),
    (7, 'Physical Infrastructure'),
    (DCIM_OTHER, 'Other Object'),
)


@python_2_unicode_compatible
class Switch(RacksModel):
    dcim_id = models.IntegerField(primary_key=True)
    label = models.CharField(max_length=255)
    management_address = models.CharField(max_length=255)

    def active_ports(self):
        return [p for p in self.ports.all() if p.dcim_device_type != DCIM_EMPTY]

    def empty_ports(self):
        return [p for p in self.ports.all() if p.dcim_device_type == DCIM_EMPTY]

    def ports_to_servers(self):
        return [p for p in self.ports.all() if p.dcim_device_type == DCIM_SERVER]

    def ports_to_switches(self):
        return [p for p in self.ports.all() if p.dcim_device_type == DCIM_SWITCH]

    def ports_to_others(self):
        return [p for p in self.ports.all() if p.dcim_device_type == DCIM_OTHER]

    @cached_property
    def locations(self):
        locations = set()
        for port in self.ports.all():
            for address in port.l2addresses.all():
                if address.inventory_object and address.inventory_object.facter_location:
                    locations.add(address.inventory_object.facter_location.location)
        return sorted(locations, key=lambda l: l.name)

    @property
    def dcim_link(self):
        return settings.DCIM_URL + 'devices.php?' + urlencode({'DeviceID': self.dcim_id})

    @property
    def ssh_link(self):
        if self.management_address:
            return 'ssh://%s' % self.management_address
        return None

    def __str__(self):
        return self.label


@python_2_unicode_compatible
class SwitchPort(RacksModel):
    number = models.CharField(max_length=255, null=True)
    label = models.CharField(max_length=255, null=True)
    switch = models.ForeignKey(Switch, related_name='ports')
    l2addresses = models.ManyToManyField('inventory.L2Address', related_name='switch_ports')
    dcim_server_id = models.IntegerField(null=True)
    switch_port_name = models.CharField(max_length=255, null=True)
    dcim_device_type = models.IntegerField(default=DCIM_EMPTY, choices=DCIM_DEVICE_TYPES)

    def l2addresses_with_objects(self):
        return [a for a in self.l2addresses.all() if a.inventory_object]

    def l2addresses_without_objects(self):
        return [a for a in self.l2addresses.all() if not a.inventory_object]

    class Meta:
        unique_together = ('switch', 'number')

    def __str__(self):
        if self.label is not None:
            return u'Port "{}"'.format(self.label)
        if self.switch_port_name is not None:
            return u'Port "{}"'.format(self.switch_port_name)
        if self.number is not None:
            return u'Port #{}'.format(self.number)
        return u'Unlabeled Port (UID #{}) '.format(self.pk)


# Inheritance from generic models, because this models aren't part of application
# rather local copy of DCIM entries
@python_2_unicode_compatible
class Device(models.Model):
    # reference for models definition:
    # https://github.com/samilliken/openDCIM/blob/master/create.sql#L250-L296
    label = models.CharField(max_length=80)
    device_type = models.CharField(max_length=23, default="Server")
    install_date = models.DateField()
    owner = models.IntegerField()
    contact_person = models.ForeignKey(User, null=True)
    notes = models.TextField(null=True)
    custom_values = jsonfield.JSONField(default={})
    """Custom values ID:
    * 2    Ticket
    * 3    Price Euro
    * 4    Price Dollar
    * 5    y-m-d: Exp. date
    * 7    Project
    * 10   IPMI
    * 13   Contact Person
    * 16   Approval Manager
    * 19   Reporter
    * 22   y-m-d: Beginning of commitment period
    * 23   y-m-d: Ending of commitment period
    """
    @property
    def jira_issue(self):
        return self.custom_values.get('2', '')

    @property
    def price_euro(self):
        return self.custom_values.get('3', '')

    @property
    def price_usd(self):
        return self.custom_values.get('4', '')

    @property
    def contact_person_string(self):
        return self.custom_values.get('13', '')

    @property
    def url(self):
        if isinstance(settings.DCIM_URL, six.string_types):
            return "%s/devices.php?DeviceID=%s" % (settings.DCIM_URL.rstrip('/'), str(self.pk))
        raise ImproperlyConfigured("Wrong type of settings.DCIM_URL. DCIM_URL must be a string.")

    class Meta:
        ordering = ('label', )

    def __str__(self):
        return "%s (ID: %s)" % (self.label, self.pk)

    def __repr__(self):
        return "<DCIM Device: %s>" % self.label


@receiver(models.signals.pre_save, sender=Device)
def device_set_contact_person(sender, instance, *args, **kwargs):
    """
    This is guess-a-user function because in DCIM we have text field Contact Person
    with format free input value.
    """
    full_name = instance.contact_person_string
    if not full_name:
        return None

    if ' ' in full_name:
        first_name, last_name = full_name.split(' ', 1)
        try:
            user = User.objects.get(first_name=first_name, last_name=last_name)
        except (User.DoesNotExist, User.MultipleObjectsReturned):
            # Try shorten name
            try:
                user = User.objects.get(email__istartswith=first_name[0] + last_name)
            except (User.DoesNotExist, User.MultipleObjectsReturned):
                return None
    else:
        try:
            user = User.objects.get(email__istartswith=full_name)
        except (User.DoesNotExist, User.MultipleObjectsReturned):
            return None

    if user != instance.contact_person:
        instance.contact_person = user
