from __future__ import absolute_import

import logging
from functools import wraps

from celery import shared_task, chain, Task
from django.conf import settings

import racks.dcim.importer as dcim_importer
import racks.importers.ldap as importers_ldap
import racks.importers.switches as importers_switches
import racks.importers.zabbix as importers_zabbix
import racks.importers.puppet as importers_puppet
import racks.importers.jenkins as importers_jenkins


log = logging.getLogger(__name__)


def racks_task(func):
    @shared_task(bind=True, ignore_result=True)
    @wraps(func)
    def task(self, *args, **kwargs):
        log.info("Starting task %s...", self.name)
        try:
            func(*args, **kwargs)
        except Exception, e:
            log.exception("Task %s failed: %s.", self.name, e)
        else:
            log.info("Task %s done.", self.name)

    task.time_limit = settings.TASK_TIME_LIMITS.get(task.name, 60)

    return task


@racks_task
def import_zabbix_hosts():
    importers_zabbix.import_hosts()


@racks_task
def import_puppet_attributes():
    importers_puppet.import_puppet_attributes()


@racks_task
def import_dcim_servers():
    dcim_importer.import_servers()


@racks_task
def import_dcim_switches():
    dcim_importer.import_switches()


@racks_task
def import_switches_address_tables():
    importers_switches.import_address_tables()


@racks_task
def bind_dcim_to_switch_ports():
    importers_switches.bind_dcim_to_switch_ports()


@racks_task
def bind_dcim_servers_to_inventory():
    importers_switches.bind_dcim_servers_to_inventory()


@racks_task
def import_users_from_ldap():
    importers_ldap.import_users()


@racks_task
def prune_users_and_groups():
    importers_ldap.prune_users_and_groups()


@racks_task
def import_jenkins_task():
    importers_jenkins.import_all_jenkins()


@racks_task
def bind_jenkins_nodes_to_inventory():
    importers_jenkins.bind_jenkins_nodes_to_inventory()


@shared_task
def _chain_task_finish(name):
    log.info("Chain task %s finished." % name)


class ChainTask(Task):
    abstract = True

    def __init__(self):
        subtasks = [
            task.si() for task in self.subtasks
        ]

        subtasks.append(_chain_task_finish.si(self.name))

        self.chain = chain(*subtasks)

    def apply(self, *args, **kwargs):
        return self.chain.apply(*args, **kwargs)

    def apply_async(self, *args, **kwargs):
        return self.chain.apply_async(*args, **kwargs)


@shared_task
class import_all(ChainTask):
    subtasks = (
        import_users_from_ldap,
        import_zabbix_hosts,
        import_puppet_attributes,
        import_dcim_servers,
        import_dcim_switches,
        import_switches_address_tables,
        bind_dcim_to_switch_ports,
        bind_dcim_servers_to_inventory,
        import_jenkins_task,
        bind_jenkins_nodes_to_inventory,
        prune_users_and_groups,
    )
