from django import forms
from django.core.exceptions import ValidationError
from django.forms.utils import flatatt
from django.utils.datastructures import MultiValueDict, MergeDict
from django.utils.html import format_html
from django.utils.safestring import mark_safe

from racks.inventory.models import InventoryObject, InventoryObjectLabel
from racks.dcim.models import Device

from racks.widgets import FilteredSelectMultiple

try:
    from django.core.urlresolvers import reverse
except ImportError:
    from django.urls import reverse  # Django 1.10+


class TagsInput(forms.Widget):
    def render(self, name, value, attrs=None):
        if value is None:
            value = ''
        final_attrs = self.build_attrs(attrs, name=name)

        if final_attrs.get('readonly'):
            output = [
                '<div class="form-control" readonly="readonly">'
            ]
            for text in value:
                output.append(format_html('<span class="label label-default">{0}</span>',
                                          text))
            output.append('</div>')
        else:
            final_attrs['data-role'] = 'tagsinput'
            output = [
                format_html('<select multiple="multiple"{0}>', flatatt(final_attrs))
            ]
            for text in value:
                output.append(format_html('<option value="{0}">{1}</option>', text, text))

        output.append('</select>')
        return mark_safe('\n'.join(output))

    def value_from_datadict(self, data, files, name):
        if isinstance(data, (MultiValueDict, MergeDict)):
            return data.getlist(name)
        return data.get(name, None)


class TagsField(forms.Field):
    widget = TagsInput

    def prepare_value(self, value):
        if value is None:
            return None

        if isinstance(value, self.BoundData):
            return value.data

        objects = InventoryObjectLabel.objects.filter(pk__in=value)
        return [obj.name for obj in objects]

    class BoundData:
        def __init__(self, data):
            self.data = data

    def bound_data(self, data, initial):
        return self.BoundData(data)

    def to_python(self, value):
        objects = []

        for name in value:
            obj, created = InventoryObjectLabel.objects.get_or_create(name=name)
            objects.append(obj)

        return objects


class DCIMField(forms.IntegerField):
    def validate(self, value):
        super(DCIMField, self).validate(value)
        if not value:
            return

        try:
            Device.objects.get(pk=value)
        except:
            raise ValidationError(mark_safe(
                "DCIM object with ID '%s' not found, please check in <a href='%s'>DCIM</a>" % (
                    value,
                    Device(id=value).url
                )
            ))


class InventoryObjectForm(forms.ModelForm):
    label = TagsField(required=None)
    dcim = DCIMField(label='DCIM ID', widget=forms.TextInput(), required=False)

    def __init__(self, *args, **kwargs):
        super(InventoryObjectForm, self).__init__(*args, **kwargs)

        # DISABLED FOR NOW
        # It is optional in model due to importers, but user aren't allowed
        # to lease hostname blank
        # self.fields['name'].required = True

    def full_clean(self):
        super(InventoryObjectForm, self).full_clean()
        if 'dcim' in self.errors and 'already exists' in str(self.errors['dcim']):
            inventory_object = InventoryObject.objects.get(dcim_id=int(self.data.get('dcim')))
            self.add_error('dcim', mark_safe("Please remove <a href='%s'>%s</a>." % (
                reverse('inventory.objects.detail', kwargs={'pk': inventory_object.pk}),
                inventory_object.display_name
            )))

    class Meta:
        model = InventoryObject
        fields = [
            'name',
            'type',
            'facter_location',
            'label',
            'note',
            'l2network',
            'ipv4network',
            'ipmi_host',
            'ipmi_user',
            'ipmi_password',
            'dcim',
            'bugs',
            'assigned_user',
            'assigned_group',
        ]
        widgets = {
            'l2network': FilteredSelectMultiple,
            'ipv4network': FilteredSelectMultiple,
            'bugs': forms.TextInput(attrs={'data-role': 'bugsinput'}),
            'ipmi_user': forms.TextInput(attrs={'autocomplete': 'off'}),
            'ipmi_password': forms.TextInput(attrs={'autocomplete': 'off'}),
        }
