# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import migrations
from django.contrib.auth import management
from django.core.exceptions import ObjectDoesNotExist

def responsible2permissions(apps, schema_editor):
    "Forwards migration. Convert `responsible` relations to user permissions."
    Permission = apps.get_model('auth.Permission')
    InventoryObject = apps.get_model('inventory.InventoryObject')
    Lab = apps.get_model('inventory.Lab')
    UserObjectPermission = apps.get_model('guardian.UserObjectPermission')

    try:
        # Check if permissions have been created already
        view_ipmi = Permission.objects.get(codename='view_ipmi')
        change_io = Permission.objects.get(codename='change_inventoryobject')
        change_lab = Permission.objects.get(codename='change_lab')
    except ObjectDoesNotExist:
        # Force permissions generation for required apps
        apps.models_module = True
        management.create_permissions(apps, verbosity=0)
        apps.models_module = None
        # Attempt to get permissions again. If it fails, there's nothing we can do anyway
        view_ipmi = Permission.objects.get(codename='view_ipmi')
        change_io = Permission.objects.get(codename='change_inventoryobject')
        change_lab = Permission.objects.get(codename='change_lab')

    permissions = []

    UserObjectPermission.objects.all().delete()

    objects = InventoryObject.objects.prefetch_related('responsible')\
        .filter(responsible__isnull=False).distinct()
    for obj in objects:
        for user in obj.responsible.all():
            permissions.append(UserObjectPermission(
                permission_id=change_io.pk, content_type_id=change_io.content_type_id,
                user_id=user.pk, object_pk=str(obj.pk)
            ))
            permissions.append(UserObjectPermission(
                permission_id=view_ipmi.pk, content_type_id=view_ipmi.content_type_id,
                user_id=user.pk, object_pk=str(obj.pk)
            ))

    labs = Lab.objects.prefetch_related('responsible').filter(responsible__isnull=False).distinct()
    for lab in labs:
        for user in lab.responsible.all():
            permissions.append(UserObjectPermission(
                permission_id=change_lab.pk, content_type_id=change_lab.content_type_id,
                user_id=user.pk, object_pk=str(lab.pk)
            ))

    UserObjectPermission.objects.bulk_create(permissions)


def permissions2responsible(apps, schema_editor):
    "Backwards migration. Convert user permissions to `responsible` assignments."
    User = apps.get_model('auth.User')
    Permission = apps.get_model('auth.Permission')
    UserObjectPermission = apps.get_model('guardian.UserObjectPermission')
    InventoryObject = apps.get_model('inventory.InventoryObject')
    Lab = apps.get_model('inventory.Lab')

    change_io = Permission.objects.get(codename='change_inventoryobject')
    change_lab = Permission.objects.get(codename='change_lab')

    responsible_for_inventory = UserObjectPermission.objects.filter(
        permission_id=change_io.pk, content_type_id=change_io.content_type_id
    ).values('user_id', 'object_pk')

    responsible_for_labs = UserObjectPermission.objects.filter(
        permission_id=change_lab.pk, content_type_id=change_lab.content_type_id
    ).values('user_id', 'object_pk')

    # Aggregate data for bulk insert:
    # users = dict(user_id: tuple(list(inv_id_list), list(lab_id_list)))
    users = {}
    for resp in responsible_for_inventory:
        users.setdefault(resp['user_id'], ([], []))[0].append(int(resp['object_pk']))
    for resp in responsible_for_labs:
        users.setdefault(resp['user_id'], ([], []))[1].append(int(resp['object_pk']))

    for user_id, (inv_id_list, lab_id_list) in users.iteritems():
        user = User.objects.get(pk=user_id)
        for obj in InventoryObject.objects.filter(pk__in=inv_id_list):
            user.inventory_objects.add(obj)
        for lab in Lab.objects.filter(pk__in=lab_id_list):
            user.labs.add(lab)


class Migration(migrations.Migration):

    dependencies = [
        ('inventory', '0002_auto_20160728_1128'),
        ('guardian', '0001_initial'),
        ('auth', '0006_require_contenttypes_0002'),
        ('contenttypes', '0002_remove_content_type_name'),
    ]

    operations = [
        migrations.RunPython(responsible2permissions, permissions2responsible)
    ]
