import netaddr
import re
from django.utils.six.moves.urllib.parse import urlparse
from django.utils.encoding import python_2_unicode_compatible
from django.utils.safestring import mark_safe
import logging
import json

from collections import OrderedDict, namedtuple
from jsonfield import JSONField
from reversion import revisions as reversion

from django.conf import settings
from django.contrib.auth.models import User, Group
from django.core.urlresolvers import reverse
from django.core.exceptions import ValidationError
from django.core.validators import MaxValueValidator, MinValueValidator, URLValidator
from django.db import models
from django.utils.http import urlencode
from django.utils.html import strip_tags
from django.utils.functional import cached_property
from django.utils.translation import ugettext_lazy as _

from racks.inventory.fields import EncryptedCharField
from racks.models import RacksModel, RacksTimestampedModel

logger = logging.getLogger(__name__)


def validate_hostname(value):
    """
    Validator for FQDN hostnames.

    Django provides an excellend URL validator, but it accepts protocol, port, user/password and
    path information, so we need to handle that situation before deferring to it.

    :param value: string to validate
    :return: None, will raise ValidationError in case of failed validation
    """

    if not value:
        return value
    if ('/' in value) or (':' in value):
        raise ValidationError(_('Please only enter hostname: "www.google.com" is ok, but '
                                '"http://www.google.com" or "www.google.com/search" are not'))
    validator = URLValidator()
    try:
        validator('http://{}'.format(value))
    except ValidationError:
        if '.' in value:
            raise ValidationError(_('Please enter a valid hostname'))
        try:
            validator('http://{}.com'.format(value))
        except ValidationError:
            raise ValidationError(_('Please enter a valid hostname'))


class Location(RacksTimestampedModel):
    """
    Generic location (usually a city).
    """

    name = models.CharField(max_length=254, unique=True)

    def __unicode__(self):
        return self.name

    class Meta:
        ordering = ('name',)


class FacterLocation(RacksTimestampedModel):
    """
    Represents a specific location within a Location.

    Named so after Puppet convention to call any solid piece of information about a server "facts".

    Usually a codename for some specific place/datacenter.
    """

    fact = models.CharField(max_length=32, unique=True)
    location = models.ForeignKey('Location', on_delete=models.SET_NULL, related_name='facts',
                                 null=True, blank=True)

    def __unicode__(self):
        return "{} ({})".format(self.fact, self.location)


@reversion.register(exclude=["modified", "created"])
class L2Network(RacksTimestampedModel):
    """
    Represents a single VLAN network.

    May contain a number of IPv4 physical networks.
    """

    name = models.CharField(max_length=254)
    vlanid = models.IntegerField(validators=[
        MinValueValidator(1),
        MaxValueValidator(4094),
    ], verbose_name='VLAN ID')

    class Meta:
        verbose_name = 'VLAN'

    def __unicode__(self):
        if self.name:
            return u'{} (VLAN ID: {})'.format(self.name, self.vlanid)

        return u"VLAN ID: {}".format(self.vlanid)


@python_2_unicode_compatible
class L2Address(RacksModel):
    """
    Represents a MAC address associated with a switch port or a server.

    MAC addresses are imported from a number of sources, primarily from Zabbix and switches SSH
    interfaces. They're used to track actual connections between servers (inventory objects) and
    switches, thus forming an actual physical connections map.
    """

    MAC_ADDRESS_RE = re.compile(r'^[0-9a-f]{12}$')

    address = models.CharField(max_length=12, unique=True)
    inventory_object = models.ForeignKey('inventory.InventoryObject', null=True,
                                         related_name='l2addresses')

    @classmethod
    def is_valid_mac_address(cls, mac_address):
        return cls.MAC_ADDRESS_RE.match(mac_address)

    def __str__(self):
        return ":".join(self.address[i:i + 2] for i in range(0, len(self.address), 2)).upper()


@reversion.register(exclude=["modified", "created"])
@python_2_unicode_compatible
class IPv4Network(RacksTimestampedModel):
    """
    A single IPv4 physical or virtual network.

    IPv4 networks are provided manually by editing Inventory Objects.
    """

    address = models.CharField(max_length=15)
    netmask = models.CharField(max_length=15)
    gateway = models.CharField(max_length=15)
    l2network = models.ForeignKey(
        L2Network, on_delete=models.SET_NULL, related_name='ipv4_networks', null=True, blank=True)

    class Meta:
        unique_together = ('address', 'netmask')
        ordering = ('address',)
        verbose_name = 'IPv4 Network'

    def __str__(self):
        return str(netaddr.IPNetwork('%s/%s' % (self.address, self.netmask)))


@python_2_unicode_compatible
class InventoryObjectType(RacksModel):
    """
    Types of Inventory Objects.

    There are currently two inventory object types tracked: physical server and virtual machine.
    Special type "Unknown" is used when we don't know what kind of server we're dealing with.
    """

    name = models.CharField(max_length=254, unique=True)

    def __str__(self):
        return self.name


@reversion.register(exclude=["modified", "created"])
@python_2_unicode_compatible
class InventoryObjectLabel(RacksTimestampedModel):
    """
    A textual label assigned to an inventory object.

    Labels can be assigned to inventory objects in the Racks UI, and are primarily used to
    filter/search inventory. Labels are a purely internal model and are not present in any of
    3rd party services.

    Labels model contains support for multi-colored labels which is currently unused to prevent
    visual confusion with other objects (which are currently displayed in UI using differently
    colored tags).
    """

    TYPE_CHOICES = (
        ('GY', 'Grey'),
        ('BL', 'Blue'),
        ('GN', 'Green'),
        ('LB', 'Lightblue'),
        ('OR', 'Orange'),
        ('RD', 'Red'),
    )

    name = models.CharField(max_length=64, unique=True)
    type = models.CharField(
        max_length=2,
        choices=TYPE_CHOICES,
        default='GY'
    )

    def __str__(self):
        return self.name

    def get_bootstrap_color(self):
        colors = {
            'GY': 'default',
            'BL': 'primary',
            'GN': 'success',
            'LB': 'info',
            'OR': 'warning',
            'RD': 'danger',
        }

        return colors.get(self.type, 'default')

    def get_absolute_url(self):
        return reverse('inventory.objects.list') + '?' + urlencode({
            'label': self.name
        })


@reversion.register(
    exclude=["modified", "created"],
    follow=['labels', 'l2network', 'ipv4network'],
)
@python_2_unicode_compatible
class InventoryObject(RacksTimestampedModel):
    """
    Represents a single physical or virtual machine.

    This is the primary model for the entire Racks project. Practically everything in Racks is
    dealing with InventoryObjects in one way or another.

    Inventory Object model supports extended per-field permissions. At the moment, only IPMI fields
    are protected, though the list is scheduled to expand with the addition of budget-related
    fields.
    """

    name = models.CharField(max_length=254, unique=True, db_index=True, null=True, blank=True,
                            verbose_name='hostname', validators=[validate_hostname])
    type = models.ForeignKey(InventoryObjectType, related_name='inventory_objects', null=True,
                             blank=True, verbose_name='inventory object type')
    hardware = models.CharField(max_length=255, null=True, blank=True)
    facter_location = models.ForeignKey('FacterLocation', related_name='inventory_objects',
                                        null=True, blank=True)
    labels = models.ManyToManyField(InventoryObjectLabel, related_name='inventory_objects',
                                    blank=True)
    l2network = models.ManyToManyField(L2Network, related_name='inventory_objects', blank=True,
                                       verbose_name='VLAN')
    ipv4network = models.ManyToManyField(IPv4Network, related_name='inventory_objects', blank=True,
                                         verbose_name='IPv4 network')

    ipmi_host = models.CharField(max_length=255, unique=True, null=True, blank=True,
                                 verbose_name='IPMI hostname')
    ipmi_user = models.CharField(max_length=255, null=True, blank=True,
                                 verbose_name='IPMI username')
    ipmi_password = EncryptedCharField(max_length=255, null=True, blank=True,
                                       verbose_name='IPMI password')
    dcim = models.OneToOneField('dcim.Device', verbose_name='DCIM', null=True,
                                on_delete=models.SET_NULL, blank=True)

    # Keep sources as a dict {'source_name': 'source_key'}, for example
    # {'dcim': 1233} or {'user': 'admin'}
    source = JSONField(default={})
    bugs = models.CommaSeparatedIntegerField(blank=True, null=True, max_length=254)
    note = models.TextField(null=True, verbose_name='Notes', blank=True)

    assigned_user = models.ForeignKey(
        User, blank=True, null=True, related_name='assigned_objects', verbose_name='Contact Person',
        limit_choices_to={'id__gt': 0},
    )
    assigned_group = models.ForeignKey(
        Group, blank=True, null=True, related_name='assigned_objects', verbose_name='Contact Group')

    class Meta:
        ordering = ('-id', )
        permissions = (
            ('manage_inventoryobject', 'Can manage inventory object permissions'),
            ('view_ipmi', 'View IPMI'),
        )

    def get_history(self, view_ipmi=False):
        history = list(reversion.get_for_object(self))
        compiled_history = []

        fields = {field.name: field for field in InventoryObject._meta.get_fields()}
        if not view_ipmi:
            del fields['ipmi_host']
            del fields['ipmi_user']
            del fields['ipmi_password']
        old_new = namedtuple('old_new', ('old', 'new'))
        old_removed_new_added = namedtuple('old_removed_new_added',
                                           ('old', 'removed', 'new', 'added'))

        for new_version, old_version in zip(history[:-1], history[1:]):
            new_obj = new_version.object_version.object
            old_obj = old_version.object_version.object
            new_json = json.loads(new_version.serialized_data)[0]['fields']
            old_json = json.loads(old_version.serialized_data)[0]['fields']

            changed = []
            for name, field in fields.items():
                if isinstance(field, models.ManyToManyField):
                    old_value = sorted(old_json.get(name))
                    new_value = sorted(new_json.get(name))
                else:
                    old_value = getattr(old_obj, name)
                    new_value = getattr(new_obj, name)

                # skip same values
                if new_value == old_value:
                    continue

                # NULL and '' are also the same values
                if old_value is None and new_value == '' \
                        or (old_value == '' and new_value is None):
                    continue

                model = field.related_model
                if isinstance(field, models.ForeignKey):
                    try:
                        old_value = model.objects.get(pk=old_json.get(name))
                    except model.DoesNotExist:
                        old_value = None
                    try:
                        new_value = model.objects.get(pk=new_json.get(name))
                    except model.DoesNotExist:
                        # reversion.get_deleted(model, )
                        new_value = None
                elif isinstance(field, models.ManyToManyField):
                    old_value = list(model.objects.filter(pk__in=old_value))
                    new_value = list(model.objects.filter(pk__in=new_value))
                    removed = [i for i in old_value if i not in new_value]
                    added = [i for i in new_value if i not in old_value]
                    changed.append((field.name, old_removed_new_added(
                        old_value, removed, new_value, added)))
                    continue  # we need to skip changed.append((field.name, old_value, new_value))

                changed.append((field.name, old_new(old_value, new_value)))

            # empty list could be if changes were applied to forbidden fields
            # or if change has fields that were removed from current
            # app version and we should not add it to output
            if changed:
                compiled_history.append((
                    new_version.revision.user,
                    new_version.revision.comment,
                    new_version.revision.date_created,
                    changed,
                ))
        return compiled_history

    def is_owner(self, applicant):
        if isinstance(applicant, Group):
            return applicant == self.assigned_group
        if isinstance(applicant, User):
            return (applicant == self.assigned_user) or \
                   self.assigned_group and (applicant in self.assigned_group.user_set.all())
        return False

    def __str__(self):
        return "{} ({})".format(self.name, self.dcim_label)

    def update_attributes(self, attributes):
        for name, value in attributes.items():
            attribute_object, created = self.attributes.get_or_create(name=name,
                                                                      defaults={'value': value})

            if not created and attribute_object.value != value:
                attribute_object.value = value
                attribute_object.save()

    @cached_property
    def all_attributes(self):
        return dict((attr.name, attr.value) for attr in self.attributes.all())

    # TODO: move this info to InventoryObject
    @property
    def cpu(self):
        try:
            return '{processorcount}x{processor0}'.format(**self.all_attributes)
        except KeyError:
            return None

    @property
    def display_name(self):
        if self.name:
            return self.name
        elif self.dcim:
            return 'DCIM Entry: %s' % self.dcim.label
        else:
            return 'Empty object ID: %s' % self.id

    def dump(self):
        return {
            'name': self.name,
            'hardware': self.hardware,
        }

    @property
    def ipmi_link(self):
        if not self.ipmi_host:
            return None
        if urlparse(self.ipmi_host).scheme != '':
            return self.ipmi_host
        return 'http://%s' % self.ipmi_host

    @property
    def ipv4address(self):
        return self.all_attributes.get('ipaddress', None)

    @property
    def ipv4networks(self):
        return [
            '%s/%s' % (x.address, x.netmask) for x in self.ipv4network.all()]

    @property
    def jira_issue(self):
        issue = self.attributes.filter(name__in=['DCIM Jira Issue', 'Jira Issue', 'jira_issue'])
        if issue:
            return '%s%s' % (settings.JIRA_BROWSE_URL, issue[0].value)
        return None

    @property
    def macaddress(self):
        result = []
        for key, value in self.all_attributes.items():
            if 'macaddress' in key:
                result.append(value)
        return result

    # TODO: move this info to InventoryObject
    @property
    def memory(self):
        return self.all_attributes.get('memorysize', None)

    @property
    def ssh_link(self):
        try:
            return 'ssh://{fqdn}'.format(**self.all_attributes)
        except KeyError:
            return None

    def get_absolute_url(self):
        return reverse('inventory.objects.detail', args=(self.pk,))

    @property
    def label_string(self):
        return ','.join([x.name for x in self.labels.all()])

    @property
    def macaddress_string(self):
        return ','.join(self.macaddress)

    @property
    def assigned_user_name(self):
        if self.assigned_user is None:
            return u''
        return self.assigned_user.username

    @property
    def assigned_group_name(self):
        if self.assigned_group is None:
            return u''
        return self.assigned_group.name

    @property
    def zabbix_link(self):
        url = 'https://monitoring.infra.mirantis.net/hostinventories.php?hostid={zabbix_host_id}'
        try:
            return url.format(**self.all_attributes)
        except KeyError:
            return None

    @property
    def dcim_link(self):
        return self.dcim.url if self.dcim else ''

    @property
    def dcim_label(self):
        return self.dcim.label if self.dcim else ''

    @property
    def dcim_jira_issue(self):
        return self.dcim.jira_issue if self.dcim else ''

    @property
    def dcim_jira_issue_link(self):
        if self.dcim_jira_issue:
            return settings.JIRA_BROWSE_URL + self.dcim_jira_issue

    @property
    def dcim_price_usd(self):
        return self.dcim.price_usd if self.dcim else ''

    @property
    def dcim_price_euro(self):
        return self.dcim.price_euro if self.dcim else ''

    @property
    def location_name(self):
        if self.facter_location and self.facter_location.location:
            return self.facter_location.location.name

    @property
    def location_fact(self):
        if self.facter_location:
            return self.facter_location.fact

    @property
    def bugs_list(self):
        return self.bugs.split(',') if self.bugs else []

    @property
    def used_in_string(self):
        # DON'T USE filter here, as it used in CSV export with prefetch related
        allocs = [node.instance.name for node in self.jenkins_nodes.all() if not node.offline]
        allocs.extend([lab.name for lab in self.labs.all()])
        return ', '.join(sorted(allocs))

    @property
    def comment(self):
        comment = []
        # DON'T USE filter here, as it used in CSV export with prefetch related
        offline_nodes = [node for node in self.jenkins_nodes.all() if node.offline]
        for node in offline_nodes:
            comment.append('Offline in %s: %s' % (
                node.instance.name,
                strip_tags(node.offline_reason) if node.offline_reason else 'no comment'
            ))
        return '. '.join(comment)

    SEARCHABLE_FIELDS_MAPPING = OrderedDict([
        ('Name', 'name'),
        ('Hardware', 'hardware'),
        ('Location', 'facter_location__location__name'),
        ('Facter Location', 'facter_location__fact'),
        ('DCIM Label', 'dcim__label'),
        ('DCIM Notes', 'dcim__notes'),
        ('Type', 'type__name'),
        ('IPMI Host', 'ipmi_host'),
        ('Assigned User Username', 'assigned_user__username'),
        ('Assigned User First Name', 'assigned_user__first_name'),
        ('Assigned User Last Name', 'assigned_user__last_name'),
        ('Assigned User Email', 'assigned_user__email'),
        ('Assigned Group', 'assigned_group__name'),
    ])


@python_2_unicode_compatible
class InventoryObjectCredential(RacksTimestampedModel):
    """
    Unused. Possibly deprecated.

    This model was planned to contain various sensitive credentials for inventory object, but is
    currently unused. Credentials protection meanwhile has been implemented with Racks permission
    system ('security.permissions').
    """

    name = models.CharField(max_length=255)
    context = models.CharField(max_length=255)
    value = models.CharField(max_length=255)
    inventory_object = models.ForeignKey(InventoryObject, related_name='inventory_objects')

    class Meta:
        unique_together = ('name', 'context', 'inventory_object')

    def __repr__(self):
        return "<%s: %s@%s>" % (
            self.__class__.__name__,
            self.name,
            self.inventory_object.name if self.inventory_object else 'Unknown'
        )

    def __str__(self):
        return '%s for %s@%s' % (
            self.name, self.context, self.inventory_object.name)


@python_2_unicode_compatible
class InventoryObjectAttribute(RacksTimestampedModel):
    """
    Contains various miscellaneous data pieces imported from Zabbix and/or Puppet.

    This model is used to support the dynamic attributes for Inventory Object model.
    """

    name = models.CharField(max_length=254)
    value = models.TextField()
    object = models.ForeignKey(InventoryObject, related_name='attributes')

    class Meta:
        unique_together = ('name', 'object')
        ordering = ('name',)

    def __repr__(self):
        return "<%s: %s@%s>" % (
            self.__class__.__name__,
            self.name,
            self.object.name if self.object else 'Unknown'
        )

    def __str__(self):
        return self.name

    @property
    def html_value(self):
        handlers = [
            'fqdn',
            'version',
            'zabbix_host_id',
        ]
        value = None
        if self.name in handlers:
            if self.name == 'zabbix_host_id':
                value = '<a href="%s" target="_blank">%s</a>' % (
                    self.object.zabbix_link,
                    self.value
                )
            if self.name == 'version' and settings.COMMIT_HASH_URL_PATTERN:
                sha1commit = re.search(r'([a-f0-9]{40})', self.value)
                if sha1commit and len(sha1commit.groups()) > 0:
                    value = '<a href="%s" target="_blank">%s</a>' % (
                        settings.COMMIT_HASH_URL_PATTERN % (sha1commit.groups()[0]),
                        self.value,
                    )
            if self.name == 'fqdn':
                value = '<a href="ssh://%s">%s</a>' % (
                    self.value,
                    self.value,
                )
        return mark_safe(value) if value else self.value


@python_2_unicode_compatible
class Lab(RacksTimestampedModel):
    """
    Represents a single lab.

    Lab is a group of inventory objects working on a single task within a project. This model
    currently contains a lot of extra functionality which is likely to deprecate with the
    introduction of Projects.
    """

    name = models.CharField(max_length=255, blank=False)
    vlans = models.ManyToManyField('L2Network', blank=True, verbose_name='VLAN')
    inventory_objects = models.ManyToManyField('InventoryObject', related_name='labs', blank=True)
    bugs = models.CommaSeparatedIntegerField(blank=True, null=True, max_length=254)
    assigned_user = models.ForeignKey(
        User, blank=True, null=True, related_name='assigned_labs', verbose_name='Contact Person')
    assigned_group = models.ForeignKey(
        Group, blank=True, null=True, related_name='assigned_labs', verbose_name='Contact Group')

    class Meta:
        permissions = (
            ('manage_lab', 'Can manage lab permissions'),
        )

    def __repr__(self):
        return "<%s: %s>" % (self.__class__.__name__, self.name)

    def __str__(self):
        return self.name

    def is_owner(self, applicant):
        if isinstance(applicant, Group):
            return applicant == self.assigned_group
        if isinstance(applicant, User):
            return (applicant == self.assigned_user) or \
                   self.assigned_group and (applicant in self.assigned_group.user_set.all())
        return False

    @property
    def bugs_list(self):
        return self.bugs.split(',') if self.bugs else []
