import six
import json

from django.contrib.auth.models import User
from rest_framework import serializers

from racks.inventory.models import (
    FacterLocation,
    InventoryObject,
    InventoryObjectAttribute,
    InventoryObjectType,
    InventoryObjectLabel,
    IPv4Network,
)


class UserSerializer(serializers.HyperlinkedModelSerializer):
    class Meta:
        model = User
        fields = ('id', 'url', 'username', 'email',)


class IPv4NetworkSerializer(serializers.ModelSerializer):
    class Meta:
        model = IPv4Network
        fields = ('id', 'address', 'netmask', 'gateway',)


class SlugGetOrCreateRelatedField(serializers.SlugRelatedField):
    def to_internal_value(self, data):
        return self.get_queryset().model.objects.get_or_create(**{self.slug_field: data})[0]


class AttributesField(serializers.Field):
    attr_query = []

    def bind(self, field_name, parent):
        super(AttributesField, self).bind(field_name, parent)
        if self.parent.instance and not isinstance(self.parent.instance, InventoryObject):
            self.attr_query = iter(self.parent.instance)

    def to_representation(self, value):
        instance = next(self.attr_query) if self.attr_query else self.parent.instance
        return {i.name: i.value for i in instance.attributes.all()}

    def to_internal_value(self, data):
        if isinstance(data, six.string_types):
            data = json.loads(data)
        if not self.parent.instance:
            return [
                InventoryObjectAttribute(name=key, value=value)
                for key, value in data.items()
            ]
        result = []
        for key, value in data.items():
            result.append(self.parent.instance.attributes.update_or_create(name=key, defaults={
                'value': value
            })[0])
        self.parent.instance.attributes.exclude(name__in=data.keys()).delete()
        return result


class InventoryObjectSerializer(serializers.ModelSerializer):
    labels = SlugGetOrCreateRelatedField(
        many=True,
        required=False,
        slug_field='name',
        queryset=InventoryObjectLabel.objects.all()
    )
    type = SlugGetOrCreateRelatedField(
        many=False,
        required=False,
        slug_field='name',
        queryset=InventoryObjectType.objects.all()
    )
    facter_location = SlugGetOrCreateRelatedField(
        many=False,
        required=False,
        slug_field='fact',
        queryset=FacterLocation.objects.all(),
    )
    attributes = AttributesField(
        required=False,
        allow_null=True,
    )

    def __init__(self, *args, **kwargs):
        ipmi = kwargs.pop('ipmi', False)
        super(InventoryObjectSerializer, self).__init__(*args, **kwargs)
        if not ipmi:
            del self.fields['ipmi_host']
            del self.fields['ipmi_user']
            del self.fields['ipmi_password']

    class Meta:
        model = InventoryObject
        fields = (
            'id',
            'name',
            'type',
            'facter_location',
            'ipmi_host',
            'ipmi_user',
            'ipmi_password',
            'labels',
            'attributes',
        )
