from __future__ import unicode_literals

import unittest
import json

from django.contrib.auth.models import User, Group
from rest_framework.test import APIClient, APITestCase

from racks.inventory.models import (
    InventoryObject,
    InventoryObjectLabel,
    InventoryObjectType,
    IPv4Network,
    Location,
    FacterLocation,
)
from racks.security.models import Application


# TODO: This should be refactored to the same tests for super user, assigned user, assigned group
# TODO: and django guardian and separate case for no access.
class APITestCaseJson(APITestCase):
    def assertResponseJSONContains(self, response, status_code, expected):
        if response.status_code != status_code:
            message = '{} != {}\n{}'.format(response.status_code, status_code, response)
            self.fail(message)

        response_json = response.rendered_content
        actual = json.loads(response_json)

        if isinstance(expected, list):
            self.assertIsInstance(actual, list)
            expected_items = expected
            actual_items = actual
        else:
            expected_items = [expected]
            actual_items = [actual]

        # To account for additional fields, which may appear in responses,
        # delete fields which aren't present in expected result before
        # comparison.
        # Can't use assertDictContainsSubset here, because we can have several
        # items here and I want to have diff of full response on test failure.
        for actual_item, expected_item in zip(actual_items, expected_items):
            if isinstance(expected_item, dict) and isinstance(actual_item, dict):
                extra_keys = set(actual_item) - set(expected_item)
                for key in extra_keys:
                    del actual_item[key]
        self.assertEqual(actual, expected)


class APIv2TestCase(APITestCaseJson):
    def setUp(self):
        self.user = User.objects.create_superuser('user', 'user@mail', 'password')
        self.client.force_authenticate(user=self.user)

        self.user.applications.create(name='authapp1', auth_token='authtoken1')
        self.user.applications.create(name='authapp2', auth_token='authtoken2')

        self.location1 = Location.objects.create(name='location1_name')

        self.facter_location1 = self.location1.facts.create(fact='facter_location1_name')
        self.facter_location2 = self.location1.facts.create(fact='facter_location2_name')

        self.vlan = IPv4Network.objects.create(address='10.0.0.1', netmask='255.0.0.0',
                                               gateway='10.0.0.254')

        self.type = InventoryObjectType.objects.create(name='type1')

        self.label = InventoryObjectLabel.objects.create(name='label1')

        self.obj1 = obj1 = InventoryObject.objects.create(
            name='object1.com', facter_location=self.facter_location1, type=self.type)

        obj1.attributes.create(name='attribute1', value='obj1 attribute1 value'),
        obj1.attributes.create(name='attribute2', value='obj1 attribute2 value'),

        self.obj2 = obj2 = InventoryObject.objects.create(
            name='object2.com', facter_location=self.facter_location2
        )
        obj2.attributes.create(name='attribute1', value='obj2 attribute1 value')
        obj2.labels.add(self.label)

    def test_app_authentication(self):
        authenticated_client = APIClient(HTTP_X_AUTH_APPLICATION='authapp1',
                                         HTTP_X_AUTH_TOKEN='authtoken1')

        response = authenticated_client.get('/api/v2/')

        self.assertEqual(response.status_code, 200)

    def test_user_authentication(self):
        client = APIClient()

        User.objects.create_superuser('admin', 'foo@foo.com', 'admin')
        client.login(username='admin', password='admin')

        response = client.post('/api/v2/objects/', {
            'name': 'object3.com',
        })

        self.assertEqual(response.status_code, 201)

    def test_authenticated_create(self):
        authenticated_client = APIClient(HTTP_X_AUTH_APPLICATION='authapp1',
                                         HTTP_X_AUTH_TOKEN='authtoken1')

        response = authenticated_client.post('/api/v2/objects/', {'name': 'object3.com'})
        self.assertEqual(response.status_code, 201)

        response = authenticated_client.post('/api/v2/objects/', {'name': 'object4'})
        self.assertEqual(response.status_code, 201)

        response = authenticated_client.post('/api/v2/objects/', {'name': 'illegal_hostname.l'})
        self.assertEqual(response.status_code, 400)

    def test_invalid_auth_app(self):
        authenticated_client = APIClient(
            HTTP_X_AUTH_APPLICATION='invalidauthapp',
            HTTP_X_AUTH_TOKEN='authtoken1')

        response = authenticated_client.get('/api/v2/')

        self.assertEqual(response.status_code, 403)

    def test_invalid_auth_token(self):
        authenticated_client = APIClient(HTTP_X_AUTH_APPLICATION='authapp1',
                                         HTTP_X_AUTH_TOKEN='invalidauthtoken')

        response = authenticated_client.get('/api/v2/')

        self.assertEqual(response.status_code, 403)

    def test_inactive_app(self):
        authenticated_client = APIClient(HTTP_X_AUTH_APPLICATION='authapp2',
                                         HTTP_X_AUTH_TOKEN='authtoken2')

        response = authenticated_client.get('/api/v2/')

        self.assertEqual(response.status_code, 200)

        Application.objects.filter(name='authapp2').update(active=False)

        response = authenticated_client.get('/api/v2/')

        self.assertEqual(response.status_code, 403)

    def test_modify_existing_attributes(self):
        response = self.client.patch('/api/v2/objects/1/', data={
            'attributes': json.dumps({'attribute1': 'updated value'}),
        })
        self.assertResponseJSONContains(response, 200, {
            'id': 1,
            'attributes': {
                'attribute1': 'updated value',
            }
        })

    def test_create_object(self):
        response = self.client.post('/api/v2/objects/', {
            'name': 'object3.com',
        })

        self.assertResponseJSONContains(response, 201, {
            'name': 'object3.com',
            'facter_location': None,
            'type': None,
            'labels': [],
            'attributes': {},
        })

    def test_update_object(self):
        response = self.client.patch('/api/v2/objects/{}/'.format(self.obj1.pk), {
            'name': 'modified-object1.com',
        })

        self.assertEqual(response.status_code, 200)
        self.assertEqual(InventoryObject.objects.get(id=self.obj1.pk).name, 'modified-object1.com')

    def test_create_object_with_attributes(self):
        object_response = self.client.post('/api/v2/objects/', {
            'name': 'object3.com',
            'attributes': {
                'attribute1': 'obj3 attribute1 value',
            }
        }, format='json')

        object_id = object_response.data['id']

        get_response = self.client.get('/api/v2/objects/{}/'.format(object_id))

        self.assertResponseJSONContains(get_response, 200, {
            'id': object_id,
            'name': 'object3.com',
            'facter_location': None,
            'type': None,
            'labels': [],
            'attributes': {
                'attribute1': 'obj3 attribute1 value',
            },
        })

    def test_objects_name_filter(self):
        response = self.client.get('/api/v2/objects/?name=object2.com')

        self.assertResponseJSONContains(response, 200, [
            {'id': self.obj2.pk},
        ])

    def test_patch_object_labels_format_json(self):
        patch_response = self.client.patch('/api/v2/objects/2/', {
            'labels': ['newlabel1', 'newlabel2', 'newlabel3']
        }, format='json')
        self.assertEqual(patch_response.status_code, 200)

        response = self.client.get('/api/v2/objects/{}/'.format(self.obj2.pk))
        self.assertResponseJSONContains(response, 200, {
            'id': self.obj2.pk,
            'name': 'object2.com',
            'labels': ['newlabel1', 'newlabel2', 'newlabel3'],
        })

    def test_patch_object_location(self):
        response = self.client.patch('/api/v2/objects/{}/'.format(self.obj2.pk), {
            'facter_location': 'newfact'
        }, format='json')

        self.assertEqual(response.status_code, 200)

        obj = InventoryObject.objects.get(id=self.obj2.pk)

        self.assertEqual(obj.facter_location.fact, 'newfact')

    def test_patch_object_location_to_existing_fact(self):
        response = self.client.patch('/api/v2/objects/{}/'.format(self.obj2.pk), {
            'facter_location': 'facter_location1_name'
        }, format='json')

        self.assertEqual(response.status_code, 200)

        obj = InventoryObject.objects.get(id=self.obj2.pk)
        self.assertEqual(obj.facter_location.fact, 'facter_location1_name')


class GenericAPITests(object):
    maxDiff = None

    def setUp(self):
        self.vlan = IPv4Network.objects.create(address='10.0.0.1', netmask='255.0.0.0',
                                               gateway='10.0.0.254')
        self.obj2 = InventoryObject.objects.create(
            name='object2.com', facter_location=FacterLocation.objects.create(fact='location2')
        )
        self.obj2.attributes.create(name='attribute1', value='obj2 attribute1 value')
        self.label = InventoryObjectLabel.objects.create(name='label1')
        self.obj2.labels.add(self.label)

    def test_get_objects(self):
        response = self.client.get('/api/v2/objects/?ordering=id')
        self.assertResponseJSONContains(response, 200, [
            {
                'id': self.obj.pk,
                'name': self.obj.name,
                'facter_location':
                    self.obj.facter_location.fact if self.obj.facter_location else None,
                'type': self.obj.type.name if self.obj.type else None,
                'labels': [label.name for label in self.obj.labels.all()],
                'attributes': {
                    attr.name: attr.value
                    for attr in self.obj.attributes.all()
                },
                # TODO: Add check for following condition:
                # Those fields are available only for superuser
                # 'ipmi_host': self.obj.ipmi_host,
                # 'ipmi_user': self.obj.ipmi_user,
                # 'ipmi_password': self.obj.ipmi_password,
            },
            {
                'id': self.obj2.pk,
                'name': self.obj2.name,
                'facter_location':
                    self.obj2.facter_location.fact if self.obj.facter_location else None,
                'type': self.obj2.type.name if self.obj2.type else None,
                'labels': [label.name for label in self.obj2.labels.all()],
                'attributes': {
                    attr.name: attr.value
                    for attr in self.obj2.attributes.all()
                },
                # TODO: Add check for following condition:
                # Those fields are available only for superuser
                # 'ipmi_host': self.obj2.ipmi_host,
                # 'ipmi_user': self.obj2.ipmi_user,
                # 'ipmi_password': self.obj2.ipmi_password,
            },
        ])

    def test_objects_location_fact_filter(self):
        location_fact = self.obj.facter_location.fact
        response = self.client.get('/api/v2/objects/?location__fact=%s' % location_fact)
        self.assertResponseJSONContains(response, 200, [{'id': self.obj.id}])

    def test_object_type_name_filter(self):
        type_name = self.obj.type.name
        response = self.client.get('/api/v2/objects/?type__name=%s' % type_name)

        self.assertResponseJSONContains(response, 200, [{'id': self.obj.pk}])

    def test_get_users(self):
        response = self.client.get('/api/v2/users/')
        self.assertResponseJSONContains(response, 200, [{
                'id': 1,
                'url': 'http://testserver/api/v2/users/1/',
                'username': 'AnonymousUser',
                'email': '',
            },
            {
                'id': 2,
                'url': 'http://testserver/api/v2/users/2/',
                'username': 'user1',
                'email': '',
            },
        ])

    def test_get_ipv4networks(self):
        response = self.client.get('/api/v2/ipv4networks/')
        self.assertResponseJSONContains(response, 200, [{
            'id': self.vlan.pk,
            'address': self.vlan.address,
            'netmask': self.vlan.netmask,
            'gateway': self.vlan.gateway,
        }])

    def test_get_ipmi_parameters(self):
        response = self.client.get('/api/v2/objects/{}/'.format(self.obj.pk))
        self.assertResponseJSONContains(response, 200, {
            'ipmi_host': 'ipmi_host',  'ipmi_user': 'ipmi_user', 'ipmi_password': 'ipmi_password',
            'facter_location': 'facter_location', 'type': 'object_type'
        })

    def test_update_ipmi_parameters(self):
        self.client.patch('/api/v2/objects/{}/'.format(self.obj.pk), {
            'ipmi_host': 'host1',  'ipmi_user': 'user1', 'ipmi_password': 'password1'
        }, format='json')
        response = self.client.get('/api/v2/objects/{}/'.format(self.obj.pk))
        self.assertResponseJSONContains(response, 200, {
            'ipmi_host': 'host1',  'ipmi_user': 'user1', 'ipmi_password': 'password1'
        })

    def test_update_properties(self):
        self.client.patch('/api/v2/objects/{}/'.format(self.obj.pk), {
            'name': 'updated.com'
        }, format='json')
        response = self.client.get('/api/v2/objects/{}/'.format(self.obj.pk))
        self.assertResponseJSONContains(response, 200, {
            'name': 'updated.com'
        })

    def test_update_attributes(self):
        self.client.patch('/api/v2/objects/{}/'.format(self.obj.pk), {
            'attributes': {'attribute1': 'updated attribute1', 'new': 'attribute'}
        }, format='json')
        response = self.client.get('/api/v2/objects/{}/'.format(self.obj.pk))
        self.assertResponseJSONContains(response, 200, {
            'attributes': {'attribute1': 'updated attribute1', 'new': 'attribute'}
        })

    def test_update_facter_location(self):
        self.client.patch('/api/v2/objects/{}/'.format(self.obj.pk), {
            'facter_location': 'new_location'
        })
        response = self.client.get('/api/v2/objects/{}/'.format(self.obj.pk))
        self.assertResponseJSONContains(response, 200, {
            'facter_location': 'new_location'
        })

    def test_update_type(self):
        self.client.patch('/api/v2/objects/{}/'.format(self.obj.pk), {
            'type': 'new_type'
        })
        response = self.client.get('/api/v2/objects/{}/'.format(self.obj.pk))
        self.assertResponseJSONContains(response, 200, {
            'type': 'new_type'
        })

    def test_update_labels(self):
        self.client.patch('/api/v2/objects/{}/'.format(self.obj.pk), {
            'labels': ['label1', 'label2']
        })
        response = self.client.get('/api/v2/objects/{}/'.format(self.obj.pk))
        self.assertResponseJSONContains(response, 200, {
            'labels': ['label1', 'label2']
        })

    def test_delete_inventory_object(self):
        response = self.client.delete('/api/v2/objects/{}/'.format(self.obj.pk))
        self.assertEqual(response.status_code, 204)
        response = self.client.get('/api/v2/objects/{}/'.format(self.obj.pk))
        self.assertEqual(response.status_code, 404)


class AssignedUserTestCase(GenericAPITests, APITestCaseJson):
    def setUp(self):
        self.user = User.objects.create_user(username='user1', password='password1')
        self.client.force_authenticate(user=self.user)
        self.obj = InventoryObject.objects.create(
            name='assigned-group.com', assigned_user=self.user,
            ipmi_user='ipmi_user', ipmi_host='ipmi_host',
            ipmi_password='ipmi_password',
            facter_location=FacterLocation.objects.create(fact='facter_location'),
            type=InventoryObjectType.objects.create(name='object_type'),
        )
        self.attrs = [
            self.obj.attributes.create(name='attribute1', value='obj1 attribute1'),
            self.obj.attributes.create(name='attribute2', value='obj1 attribute2'),
        ]
        super(AssignedUserTestCase, self).setUp()


class AssignedGroupTestCase(GenericAPITests, APITestCaseJson):
    def setUp(self):
        self.user = User.objects.create_user(username='user1', password='password1')
        self.client.force_authenticate(user=self.user)
        self.group = Group.objects.create(name='group')
        self.group.user_set.add(self.user)
        self.obj = InventoryObject.objects.create(
            name='assigned-group.com', assigned_group=self.group,
            ipmi_user='ipmi_user', ipmi_host='ipmi_host',
            ipmi_password='ipmi_password',
            facter_location=FacterLocation.objects.create(fact='facter_location'),
            type=InventoryObjectType.objects.create(name='object_type'),
        )
        self.attrs = [
            self.obj.attributes.create(name='attribute1', value='obj1 attribute1'),
            self.obj.attributes.create(name='attribute2', value='obj1 attribute2'),
        ]
        super(AssignedGroupTestCase, self).setUp()


@unittest.skip("TODO: Fix unprivileged test cases")
class UnprivilegedUserTestCase(APITestCaseJson):
    def setUp(self):
        user1 = User.objects.create_user(username='user1', password='password1')
        self.client.force_authenticate(user=user1)

        self.obj1 = obj1 = InventoryObject.objects.create(name='object1.com')
        self.attr1 = obj1.attributes.create(name='attribute1', value='obj1 attribute1 value')
        self.attr2 = obj1.attributes.create(name='attribute2', value='obj1 attribute2 value')

        self.label = InventoryObjectLabel.objects.create(name='label1')

        self.obj1.labels.add(self.label)

    def test_hidden_values(self):
        response = self.client.get('/api/v2/objects/{}/'.format(self.obj1.pk))
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.data, {
            "id": self.obj1.pk,
            "name": "object1.com",
            "location": None,
            "type": None,
            "labels": ["label1"],
            "attributes": {
                "attribute1": "obj1 attribute1 value",
                "attribute2": "obj1 attribute2 value",
            },
        })

    def test_assigned_user(self):
        response = self.client.get('/api/v2/objects/{}/'.format(self.obj1.pk))
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.data, {
            "id": self.obj1.pk,
            "name": "object1.com",
            "location": None,
            "type": None,
            "labels": ["label1"],
            "attributes": {
                "attribute1": "obj1 attribute1 value",
                "attribute2": "obj1 attribute2 value",
            },
        })

    def test_related_field_partial_update(self):
        response = self.client.patch('/api/v2/objects/{}/'.format(self.obj1.pk), {
            'name': 'newname.com',
            'labels': ['newlabel1', 'newlabel2'],
        }, format='json')

        self.assertEqual(response.status_code, 200)

        obj1 = InventoryObject.objects.get(id=self.obj1.pk)

        self.assertEqual(obj1.name, 'newname.com')

        self.assertItemsEqual(obj1.label.values_list('name', flat=True),
                              ['newlabel1', 'newlabel2'])

    def test_related_field_partial_update_no_permission(self):
        response = self.client.patch('/api/v2/objects/{}/'.format(self.obj1.pk), {
            'name': 'newname.com',
            'labels': ['newlabel1', 'newlabel2'],
        }, format='json')

        self.assertEqual(response.status_code, 200)

        obj1 = InventoryObject.objects.get(id=self.obj1.pk)

        self.assertEqual(obj1.name, 'newname.com')

        self.assertItemsEqual(obj1.label.values_list('name', flat=True),
                              ['label1'])
