
"""
This file demonstrates writing tests using the unittest module. These will pass
when you run "manage.py test".

Replace this with more appropriate tests for your application.
"""

from urllib import quote_plus

from django.contrib.auth.models import User, Group
from django.core.urlresolvers import reverse
from django.conf import settings
from django.test import TestCase, Client

from racks.inventory.models import (
    InventoryObject,
    IPv4Network,
    L2Network,
    Lab,
    Location,
)
from racks.dcim.models import Device

from guardian.shortcuts import assign_perm


class BaseTestCase(TestCase):
    def assert_object_in_search_response_context(self, query, assert_in=True):
        method = self.assertIn if assert_in else self.assertNotIn
        response = self.client.get(
            reverse('inventory.objects.search'), {'query': query})
        method(self.io.display_name, response.context['attributes'])


class WebUITestCase(TestCase):
    maxDiff = None

    def setUp(self):
        self.client = Client()

        InventoryObject.objects.create(
            id=1,
            name='object1',
        )

        L2Network.objects.create(id=1, vlanid=1)

        Lab.objects.create(id=1)

        Location.objects.create(id=1)

        IPv4Network.objects.create(id=1, address="1.2.3.4", netmask="25")

        user = User.objects.create(username='usual_user')
        self.urls = [
            '/',
            '/?label=mylabel',
            '/new',
            '/1',
            '/1/edit',

            '/vlans/',
            '/vlans/new',
            '/vlans/1/edit',

            '/labs/',
            '/labs/new',
            '/labs/1/edit',

            '/locations/',
            '/locations/new',
            '/locations/1/edit',

            '/ipv4networks/',
            '/ipv4networks/1',
            '/ipv4networks/1/edit',

            '/labels/',
            '/search.html?query=blah',

            '/account/%s/assigned' % user.username,
        ]

    def test_authenticated_access(self):
        User.objects.create_superuser(username='testuser', password='testpassword',
                                      email='hello@world')

        self.client.login(username='testuser', password='testpassword')

        responses = {
            url: self.client.get(url)
            for url in self.urls
        }

        self.assertEqual({
            url: {
                'status_code': response.status_code,
                'content_type': response['Content-Type']
            }
            for url, response in responses.items()
        }, {
            url: {
                'status_code': 200,
                'content_type': 'text/html; charset=utf-8'
            }
            for url in self.urls
        })

    def test_unautenticated_access(self):
        responses = {
            url: self.client.get(url)
            for url in self.urls
        }

        self.assertEqual({
            url: {
                'status_code': response.status_code,
                'location': response.get('Location')
            }
            for url, response in responses.items()
        }, {
            url: {
                'status_code': 302,
                'location': 'http://testserver%s?next=%s' % (
                    settings.LOGIN_URL,
                    quote_plus(url, safe='/')
                ),
            }
            for url in self.urls
        })


class NonPrivilegedUserTestCase(BaseTestCase):
    maxDiff = None

    def setUp(self):
        User.objects.create_user(username='testuser', password='testpassword', email='hello@world')
        self.client = Client()
        self.client.login(username='testuser', password='testpassword')

        InventoryObject.objects.create(
            id=1,
            name='object1',
        )

        L2Network.objects.create(id=1, vlanid=1)

        Lab.objects.create(id=1)

        Location.objects.create(id=1)

        IPv4Network.objects.create(id=1, address="1.2.3.4", netmask="25")

        self.urls = [
            '/new',
            '/1/edit',
            '/1/delete',
            '/vlans/new',
            '/vlans/1/edit',
            '/vlans/1/delete',
            '/labs/new',
            '/labs/1/edit',
            '/labs/1/delete',
            '/locations/new',
            '/locations/1/edit',
            '/locations/1/delete',
        ]

    def test_get(self):
        urls_with_status_codes = {
            '/new': 403,
            '/1': 200,
            '/1/edit': 403,
            '/1/delete': 403,
            '/vlans/new': 403,
            '/vlans/1/edit': 403,
            '/vlans/1/delete': 403,
            '/labs/new': 403,
            '/labs/1': 200,
            '/labs/1/edit': 403,
            '/labs/1/delete': 403,
            '/locations/new': 403,
            '/locations/1/edit': 403,
            '/locations/1/delete': 403,
        }

        status_codes = {
            url: self.client.get(url).status_code for url in urls_with_status_codes
        }

        self.assertEqual(status_codes, urls_with_status_codes)

    def test_post(self):
        # All post requests are forbidden for non-privileged user
        urls = [
            '/new',
            '/1/edit',
            '/1/delete',
            '/vlans/new',
            '/vlans/1/edit',
            '/vlans/1/delete',
            '/labs/new',
            '/labs/1/edit',
            '/labs/1/delete',
            '/locations/new',
            '/locations/1/edit',
            '/locations/1/delete',
        ]
        status_codes = {
            url: self.client.post(url).status_code for url in urls
        }

        self.assertEqual(status_codes, {url: 403 for url in urls})

    def test_redirect_on_name_match(self):
        io = InventoryObject.objects.first()
        response = self.client.get(
            reverse('inventory.objects.search'), {'query': io.name})
        self.assertRedirects(response, reverse('inventory.objects.detail', kwargs={'pk': io.pk}))

    def test_user_could_search_by_attributes(self):
        username = 'simple'
        email = 'simple@world'
        group_name = 'simple_group'
        obj_name = 'test_name'
        obj_label = 'obj_label'
        obj_jira = 'obj_jira'

        # logged in as `testuser` but InventoryObject belongs to `simple`
        user = User.objects.create_user(username=username, password='testpassword', email=email)
        group = Group.objects.create(name=group_name)
        dcim = Device.objects.create(label=obj_label, install_date='1970-01-01', owner=1)

        self.io = InventoryObject.objects.create(
            id=2,
            name=obj_name,
            dcim=dcim,
            assigned_user=user,
            assigned_group=group,
        )
        self.io.attributes.create(name='Jira Issue', value=obj_jira)

        self.assert_object_in_search_response_context(obj_name[1:])
        self.assert_object_in_search_response_context(obj_label)
        self.assert_object_in_search_response_context(obj_jira)
        self.assert_object_in_search_response_context(username)
        self.assert_object_in_search_response_context(email)
        self.assert_object_in_search_response_context(group_name)


class HiddenValuesTestCase(BaseTestCase):
    maxDiff = None

    url = '/1'

    def setUp(self):
        User.objects.create_user(username='testuser', password='testpassword')
        user2 = User.objects.create_user(username='user2', password='testpassword')
        user3 = User.objects.create_user(username='user3', password='testpassword')
        user4 = User.objects.create_user(username='user4', password='testpassword')
        user5 = User.objects.create_user(username='user5', password='testpassword')
        User.objects.create_superuser(username='superuser', password='testpassword', email='')

        self.client = Client()

        group3 = Group.objects.create(name='group3')
        group3.user_set.add(user3)

        self.io = InventoryObject.objects.create(
            id=1,
            name='name_value',
            ipmi_host='ipmi_host_value',
            ipmi_user='ipmi_user_value',
            assigned_user=user2,
            assigned_group=group3,
        )

        group5 = Group.objects.create(name='group5')
        group5.user_set.add(user5)

        self.io.labs.create(
            assigned_user=user4,
            assigned_group=group5,
        )

    def test_hidden_fields_for_privileged_user(self):
        self.client.login(username='superuser', password='testpassword')

        self.assert_fields_visible('/1')
        self.assert_fields_visible('/1/edit')

    def test_hidden_fields_for_non_privileged_user(self):
        self.client.login(username='testuser', password='testpassword')

        self.assert_fields_hidden('/1')

    def test_hidden_fields_for_assigned_user(self):
        self.client.login(username='user2', password='testpassword')

        self.assert_fields_visible('/1')
        self.assert_fields_visible('/1/edit')

    def test_hidden_fields_for_assigned_group(self):
        self.client.login(username='user3', password='testpassword')

        self.assert_fields_visible('/1')
        self.assert_fields_visible('/1/edit')

    def test_privileged_user_could_search_by_hidden_fields(self):
        self.client.login(username='user2', password='testpassword')
        self.assert_hidden_fields_searchable()

    def assert_hidden_fields_searchable(self):
        self.assert_object_in_search_response_context('ipmi_host_value')

    def assert_fields_visible(self, url):
        response = self.client.get(url)

        html = response.content

        # This should be visible for all users
        self.assertIn('name_value', html)

        self.assertIn('ipmi_host_value', html)
        self.assertIn('ipmi_user_value', html)

    def assert_fields_hidden(self, url):
        response = self.client.get(url)

        html = response.content

        # This should be visible for all users
        self.assertIn('name_value', html)

        self.assertNotIn('ipmi_host_value', html)
        self.assertNotIn('ipmi_user_value', html)


class PartialEditTestCase(TestCase):

    def setUp(self):
        self.client = Client()

        user = User.objects.create_user(username='testuser', password='testpassword',
                                        email='hello@world')
        User.objects.create_superuser(username='testsuperuser', password='testpassword',
                                      email='hello@world')

        invobj = InventoryObject.objects.create(
            id=1,
            name='object1.com',
            ipmi_host='object1.com',
        )

        assign_perm('inventory.change_inventoryobject', user, invobj)

    def test_unprivileged(self):
        self.client.login(username='testuser', password='testpassword')

        response = self.client.post('/1/edit', {'ipmi_host': 'newname', 'name': 'newname'})

        self.assertEqual(response.status_code, 302)

        modified_object = InventoryObject.objects.get(id=1)

        self.assertEqual(modified_object.name, 'newname')
        self.assertEqual(modified_object.ipmi_host, 'newname')

    def test_superuser(self):
        self.client.login(username='testsuperuser', password='testpassword')

        response = self.client.post('/1/edit', {'ipmi_host': 'newname.com', 'name': 'newname.com'})
        self.assertEqual(response.status_code, 302)
        modified_object = InventoryObject.objects.get(id=1)
        self.assertEqual(modified_object.name, 'newname.com')
        self.assertEqual(modified_object.ipmi_host, 'newname.com')

        response = self.client.post('/1/edit', {'name': 'newname'})
        self.assertEqual(response.status_code, 302)
        modified_object = InventoryObject.objects.get(id=1)
        self.assertEqual(modified_object.name, 'newname')

        response = self.client.post('/1/edit', {'name': 'illegal_newname.l'})
        self.assertEqual(response.status_code, 200)  # Remain on form page, no redirect
        modified_object = InventoryObject.objects.get(id=1)
        self.assertEqual(modified_object.name, 'newname')  # name unchanged
