from django.contrib.auth.models import User
import django_filters
from rest_framework import filters, viewsets
from guardian.shortcuts import get_perms
from rest_framework import permissions

from racks.inventory.models import (
    InventoryObject,
    IPv4Network,
)

from racks.inventory.serializers import UserSerializer
from racks.inventory.serializers import InventoryObjectSerializer
from racks.inventory.serializers import IPv4NetworkSerializer


class ReadOnlyPermission(permissions.BasePermission):
    def has_object_permission(self, request, view, obj):
        if request in permissions.SAFE_METHODS:
            return True


class UserViewSet(viewsets.ModelViewSet):
    queryset = User.objects.all()
    serializer_class = UserSerializer
    permission_classes = (ReadOnlyPermission, )
    filter_backends = (filters.DjangoFilterBackend, filters.OrderingFilter)
    filter_fields = ('id', 'username', 'email')
    ordering_fields = ('id',)
    ordering = ('username', )


class InventoryObjectFilter(filters.FilterSet):
    location__fact = django_filters.CharFilter(name='facter_location__fact')

    class Meta:
        model = InventoryObject
        fields = ['name', 'location__fact', 'type__name']


class InventoryObjectPermission(permissions.BasePermission):
    def has_object_permission(self, request, view, obj):
        if request.method == 'GET':
            return True
        perms = get_perms(request.user, obj)
        if (request.method == 'PATCH' and 'change_inventoryobject' in perms) or \
                (request.method == 'DELETE' and 'delete_inventoryobject' in perms):
            return True
        if obj.assigned_user == request.user or \
                obj.assigned_group and request.user in obj.assigned_group.user_set.all():
            return True
        return False


class InventoryObjectViewSet(viewsets.ModelViewSet):
    queryset = (InventoryObject.objects.all()
                .order_by('-created')
                .select_related('facter_location', 'facter_location__location'))
    serializer_class = InventoryObjectSerializer
    filter_backends = (filters.DjangoFilterBackend, filters.OrderingFilter)
    filter_class = InventoryObjectFilter
    ordering_fields = ('id', )
    permission_classes = (permissions.IsAuthenticated, InventoryObjectPermission)

    def get_serializer(self, *args, **kwargs):
        if 'pk' in self.kwargs:
            obj = self.get_object()
            ipmi = 'view_ipmi' in get_perms(self.request.user, obj) or \
                   obj.assigned_user == self.request.user or \
                   obj.assigned_group and self.request.user in obj.assigned_group.user_set.all()
        elif self.request.user.is_superuser:
            ipmi = True
        else:
            ipmi = False

        serializer_class = self.get_serializer_class()
        kwargs['context'] = self.get_serializer_context()

        return serializer_class(ipmi=ipmi, *args,  **kwargs)


class IPv4NetworkViewSet(viewsets.ModelViewSet):
    queryset = IPv4Network.objects.all().order_by('-created')
    serializer_class = IPv4NetworkSerializer
    permission_classes = (ReadOnlyPermission, )
