import unicodecsv as csv
from collections import OrderedDict

from django.core.exceptions import ObjectDoesNotExist
from django.core.urlresolvers import reverse
from django.contrib.auth.decorators import login_required
from django.shortcuts import render
from django.shortcuts import HttpResponse, HttpResponseRedirect

from racks.inventory.models import (
    InventoryObject,
    InventoryObjectAttribute,
    InventoryObjectLabel,
)


@login_required
def search(request):
    query = request.GET.get('query').strip()

    try:
        return HttpResponseRedirect(reverse('inventory.objects.detail', kwargs={
            'pk': InventoryObject.objects.get(name=query).pk}))
    except InventoryObject.DoesNotExist:
        pass

    if query.startswith('^') and query.endswith('$'):
        search_type = 'iexact'
        query = query[1:-1]
    elif query.startswith('^'):
        search_type = 'istartswith'
        query = query[1:]
    elif query.endswith('$'):
        search_type = 'iendswith'
        query = query[:-1]
    else:
        search_type = 'icontains'

    attributes = _search_for_inventory_objects(search_type, query)
    object_attrs = _search_for_inventory_object_attributes(search_type, query)

    for name in object_attrs:
        if name in attributes:
            attributes[name]['results'] += object_attrs[name]['results']
        else:
            attributes[name] = object_attrs[name]

    return render(request, 'inventory/search.html', {
        'attributes': attributes,
        'query': query,
    })


@login_required
def export_object_inventory(request):
    label = request.GET.get('label')

    filters = {}
    if label is not None:
        filters['label'] = InventoryObjectLabel.objects.get(name=label)

    object_list = InventoryObject.objects.filter(**filters).select_related(
        'facter_location__location', 'type', 'assigned_user', 'assigned_group',
    ).prefetch_related(
        'labels', 'labs', 'jenkins_nodes', 'jenkins_nodes__instance', 'dcim',
    )

    columns = OrderedDict([
        ('name', 'Name'),
        ('dcim_label', 'DCIM Label'),
        ('location_name', 'Location'),
        ('location_fact', 'Location Fact'),
        ('hardware', 'Hardware'),
        ('type', 'Type'),
        ('label_string', 'Labels'),
        # TODO: optimize code for jira issue field
        # ('jira_issue', 'Jira Issue'),
        ('dcim_jira_issue', 'DCIM Jira Issue'),
        ('dcim_price_usd', 'Price USD'),
        ('dcim_price_euro', 'Price EURO'),
        ('assigned_user_name', 'Assigned User'),
        ('assigned_user__profile__department', 'Department'),
        ('assigned_group_name', 'Assigned Group'),
        ('used_in_string', 'Used In'),
        ('comment', 'Comment'),
        ('note', 'Notes'),
    ])

    response = HttpResponse(
        content_type='text/csv',
        reason='Ok',
    )

    writer = csv.writer(response)
    writer.writerow(columns.values())
    for i in object_list:
        row = []
        for c in columns.keys():
            value = i
            keys = c.split('__', 1)
            value = getattr(value, keys[0])
            while len(keys) == 2 and value:
                keys = keys[1].split('__', 1)
                try:
                    value = getattr(value, keys[0])
                except ObjectDoesNotExist:
                    value = None
            row.append(value or '')
        writer.writerow(row)

    return response


def _search_for_inventory_objects(search_type, query):
    fields_mapping = InventoryObject.SEARCHABLE_FIELDS_MAPPING.items()
    attributes = {}

    for display_name, attr_name in fields_mapping:
        search_query = {'{}__{}'.format(attr_name, search_type): query}
        inventory_objs = [
            obj for obj in InventoryObject.objects.filter(**search_query)
        ]

        for inventory_object in inventory_objs:

            value = inventory_object
            for child in attr_name.split('__'):
                value = getattr(value, child)

            data = {
                'name': display_name,
                'value': value,
            }
            name = inventory_object.display_name
            if name in attributes:
                attributes[name]['results'].append(data)
            else:
                attributes[name] = {
                    'object': inventory_object,
                    'results': [data],
                }
    return attributes


def _search_for_inventory_object_attributes(search_type, query):
    search_query = {'value__{}'.format(search_type): query}
    attributes = {}

    inventory_attributes_objs = InventoryObjectAttribute.objects.\
        filter(**search_query).select_related('object')

    for attribute in inventory_attributes_objs:
        _obj = attribute.object

        attributes[_obj.display_name] = {
            'object': _obj,
            'results': [{
                'name': attribute.name,
                'value': attribute.html_value,
            }],
        }

    return attributes
