import logging

from django.db import models
from django.contrib.auth.models import User, Group
from django.utils.encoding import python_2_unicode_compatible

logger = logging.getLogger('racks')


SOURCE_LOCAL = 0
SOURCE_LDAP = 1
SOURCES_CHOICES = (
    (SOURCE_LOCAL, 'Local'),
    (SOURCE_LDAP, 'LDAP')
)


class RacksModel(models.Model):
    """
    Ancestor to all Racks models, used when some changes to Meta must be propagated everywhere.
    """

    class Meta:
        abstract = True


class RacksTimestampedModel(RacksModel):
    """
    Base class for models with created and modified timestamps.
    """

    created = models.DateTimeField(auto_now_add=True)
    modified = models.DateTimeField(auto_now=True)

    class Meta:
        abstract = True


@python_2_unicode_compatible
class Source(RacksModel):
    """
    Registration source for User/Group.

    Describes where the user/group is coming from. There are currently two sources for users and
    groups: local and LDAP. Racks database relies on other resources (primarily LDAP) to determine
    who can and cannot login. Any users/groups which have no sources left during data import from
    LDAP database will be considered inactive and removed if possible.
    """

    source = models.IntegerField(choices=SOURCES_CHOICES, default=SOURCE_LOCAL, primary_key=True)
    users = models.ManyToManyField(User, related_name='sources')
    groups = models.ManyToManyField(Group, related_name='sources')

    def __str__(self):
        return str(SOURCES_CHOICES[self.source][1])

    def __repr__(self):
        return "<Source: %s>" % str(SOURCES_CHOICES[self.source][1])
