from __future__ import absolute_import

import os

import yaml
from celery.schedules import crontab
from django.core.exceptions import ImproperlyConfigured


def load_yaml(filename, conf_dirs):
    """
    Loads and parses a YAML configuration file.

    :param filename: name of configuration file.
    :param conf_dirs: list of folders where the file may be located.
    :return: unparsed file (must be a dictionary).
    """

    for conf_dir in conf_dirs:
        path = os.path.join(conf_dir, filename)

        if os.path.exists(path):
            with open(path) as f:
                return yaml.safe_load(f)

    raise ImproperlyConfigured('%s not found in %r' % (filename, conf_dirs))


def process_ldap_search_dict(settings, name):
    """
    Helper function for `process_ldap_settings_dict`.

    This function will convert some LDAP settings to actual objects. It is required because LDAP
    settings are currently stored as values in YAML settings file and thus need some additional
    processing before being available to Django.

    :param settings: dictionary with parsed YAML configuration file contents.
    :param name: name of LDAP search string that must be converted to LDAPSearch object.
    :return: None, will change `settings` parameter in-place.
    """

    try:
        search = settings[name].copy()
    except KeyError:
        pass
    else:
        import ldap
        from django_auth_ldap.config import LDAPSearch

        if 'scope' in search:
            search['scope'] = getattr(ldap, search['scope'])

        settings[name] = LDAPSearch(**search)


def process_ldap_settings_dict(settings):
    """
    Processes LDAP configuration settings.

    This function will check if LDAP settings are available in the configuration file, and will
    process them as necessary, creating some objects required by `django_auth_ldap` package.

    :param settings: dictionary with parsed YAML configuration file contents.
    :return: None, will change `settings` parameter in-place.
    """

    if 'AUTH_LDAP_SERVER_URI' not in settings:
        return

    from django_auth_ldap.config import GroupOfNamesType

    settings['AUTH_LDAP_GROUP_TYPE'] = GroupOfNamesType()
    settings['AUTH_LDAP_MIRROR_GROUPS'] = True

    process_ldap_search_dict(settings, 'AUTH_LDAP_USER_SEARCH')
    process_ldap_search_dict(settings, 'AUTH_LDAP_GROUP_SEARCH')

    process_ldap_search_dict(settings, 'LDAP_USER_SEARCH')
    process_ldap_search_dict(settings, 'LDAP_GROUP_SEARCH')


def process_celery_settings(settings):
    """
    Processes Celery configuration settings.

    Only `CELERYBEAT_SCHEDULE` parameter is being handled by converting it's parameters to actual
    `crontab` objects required by `django-celery`.

    :param settings: dictionary with parsed YAML configuration file contents.
    :return: None, will change `settings` parameter in-place.
    """

    if 'CELERYBEAT_SCHEDULE' in settings:
        for params in settings['CELERYBEAT_SCHEDULE'].values():
            if 'crontab' in params['schedule']:
                params['schedule'] = crontab(**params['schedule']['crontab'])


def process_auth_settings(settings):
    """
    Handles configuration files which override authentication backends.

    This function will ensure that `django-guardian` authentication backend is present in the
    `AUTHENTICATION_BACKENDS` configuration setting even if it's been overridden by YAML
    configuration file.

    :param settings: dictionary with parsed YAML configuration file contents.
    :return: None, will change `settings` parameter in-place.
    """

    guardian_backend = 'guardian.backends.ObjectPermissionBackend'
    if 'AUTHENTICATION_BACKENDS' in settings and \
            guardian_backend not in settings['AUTHENTICATION_BACKENDS']:
        settings['AUTHENTICATION_BACKENDS'].append(guardian_backend)


def update_settings(settings, conf_path_envvar, conf_path_default, filename="settings.yaml"):
    conf_path = os.getenv(conf_path_envvar, conf_path_default)
    conf_dirs = settings['SITE_CONF_DIRS'] = conf_path.split(':')

    _site_settings_dict = load_yaml(filename, conf_dirs)
    process_ldap_settings_dict(_site_settings_dict)
    process_celery_settings(_site_settings_dict)
    process_auth_settings(_site_settings_dict)

    settings.update(_site_settings_dict)
