from django import template
from django.utils.safestring import mark_safe
from django.utils.http import urlencode
from django.core.urlresolvers import reverse
from django.contrib.auth.models import User, Group, Permission

from racks.inventory.models import (
    Lab, Location, FacterLocation, InventoryObjectLabel, L2Address, InventoryObject
)

register = template.Library()


@register.filter
def racks_tag(value, extra=None):
    """
    A unified presentation filter for Racks objects, to enforce standardized look&feel for project
    entities throughout the UI.
    """
    if isinstance(value, User):
        # Small white tag with user icon, mouseover prompt with user info, URL to user details.
        return mark_safe(
            '<a href="%s" class="btn btn-default btn-xs clearance-1" title="%s <%s>">'
            '<span class="glyphicon glyphicon-user" aria-hidden="true"></span>&nbsp;%s</a>' % (
                reverse('account.profile', kwargs={'username': value.username}),
                value.get_full_name(),
                value.email,
                value.username,
            )
        )

    if isinstance(value, Group):
        # Small blue tag with group name and funny hat icon. URL currently links to inventory object
        # search, but should be eventually changed to group details page once it's implemented.
        # If optional parameter is 'user_count', tag will include group members count.
        if extra == 'user_count':
            extra = '&nbsp;<span class="badge">%d</span>' % value.user_set.count()
        else:
            extra = ''
        return mark_safe(
            '<a class="btn btn-primary btn-xs clearance-1" href="%s?%s">'
            '<span class="glyphicon glyphicon-education" aria-hidden="true"></span>'
            '&nbsp;%s%s'
            '</a>' % (
                reverse('inventory.objects.list'),
                urlencode({'assigned_group': value.name}),
                value.name,
                extra,
            )
        )

    if isinstance(value, Lab):
        # Small white tag with blackboard icon. URL leads to lab details page.
        return mark_safe(
            '<a href="%s" class="btn btn-default btn-xs">'
            '<span class="glyphicon glyphicon-blackboard" aria-hidden="true"></span>'
            '&nbsp;%s</a>' % (
                reverse('inventory.labs.detail', kwargs={'pk': value.pk}),
                value.name,
            )
        )

    if isinstance(value, InventoryObjectLabel):
        # Small darg grey tag with tags icon. URL leads to inventory object search by label.
        return mark_safe(
            '<a href="%s?%s" class="btn btn-info btn-xs">'
            '<span class="glyphicon glyphicon-tags" aria-hidden="true"></span>&nbsp;%s</a>' % (
                reverse('inventory.objects.list'),
                urlencode({'label': value.name}),
                value.name,
            )
        )

    if isinstance(value, L2Address):
        # Highlighted MAC address in fixed-width font.
        return mark_safe('<span class="l2address">%s</span>' % value)

    if isinstance(value, (Location, FacterLocation)):
        # Small grey tag with globe icon. Optional parameter value 'error' converts color to red.
        label_class = 'danger' if extra == 'error' else 'default'
        return mark_safe(
            '<span class="label label-%s">'
            '<span class="glyphicon glyphicon-globe" aria-hidden="true"></span>&nbsp;%s</span>' % (
                label_class,
                str(value)
            )
        )

    if isinstance(value, InventoryObject):
        # Plain URL to inventory object details.
        # Consider replacing with largish (btn-sm) button tag with "hdd" icon.
        return mark_safe('<a href="%s">%s</a>' % (
            reverse('inventory.objects.detail', kwargs={'pk': value.pk}), value.display_name,
        ))

    if isinstance(value, Permission):
        # Teal tag with lock icon.
        if value.codename.endswith(value.content_type.model):
            name_len = len(value.codename)-len(value.content_type.model)-1
            value = value.codename[:name_len].lower()
        else:
            value = value.codename
        return mark_safe(
            '<span class="label label-success">'
            '<span class="glyphicon glyphicon-lock" aria-hidden="true"></span>'
            '&nbsp;%s</span>' % (value)
        )

    # Fallback if there's no type match, just display a small label with configurable color.
    return mark_safe(
        '<span class="label label-%s">%s</span>' % (extra if extra else 'default', value)
    )


@register.filter
def partition(value, chunks=2):
    length = len(value)
    chunk_size = length // chunks
    big_chunks = length % chunks
    result = []
    offset = 0
    for index in xrange(chunks):
        this_chunk_size = chunk_size + int(index < big_chunks)
        result.append(value[offset:offset+this_chunk_size])
        offset += this_chunk_size
    return result


@register.filter
def permname_for_object(perm_name, obj):
    return '%s_%s' % (perm_name, obj.__class__.__name__.lower())


@register.filter
def owns_object(user, obj):
    assigned_user = getattr(obj, 'assigned_user', None)
    assigned_group = getattr(obj, 'assigned_group', None)
    return user == assigned_user or assigned_group and user in assigned_group.user_set.all()
